# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from phonopy.structure.atoms import PhonopyAtoms
import spglib as spg
from pyiron_base import Settings

__author__ = "Osamu Waseda"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Osamu Waseda"
__email__ = "waseda@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"

s = Settings()


def analyse_phonopy_equivalent_atoms(atoms, symprec=1e-5, angle_tolerance=-1.0):
    """
    Args: (read phonopy.structure.spglib for more details)
        symprec:
            float: Symmetry search tolerance in the unit of length.
        angle_tolerance:
            float: Symmetry search tolerance in the unit of angle deg.
                If the value is negative, an internally optimized routine
                is used to judge symmetry.

    """
    s.publication_add(publication())
    positions = atoms.get_scaled_positions()
    cell = atoms.cell
    types = atoms.get_chemical_symbols()
    types = list(types)
    natom = len(types)
    positions = np.reshape(np.array(positions), (natom, 3))
    cell = np.reshape(np.array(cell), (3, 3))
    unitcell = PhonopyAtoms(symbols=types, cell=cell, scaled_positions=positions)
    ops = spg.get_symmetry(unitcell, symprec=symprec, angle_tolerance=angle_tolerance)
    return ops["equivalent_atoms"]


def publication():
    return {
        "phonopy": {
            "phonopy": {
                "journal": "Scr. Mater.",
                "year": "2015",
                "title": "First principles phonon calculations in materials science",
                "author": ["Togo, A", "Tanaka, I"],
                "pages": "1--5",
                "volume": "108",
                "month": "Nov",
            }
        }
    }



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
import ast
import logging
import numpy as np

"""
Simple python executable that resembles the behavior of a real job
"""

__author__ = "Joerg Neugebauer"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


# Set the logging behaviour of the executable
logging.basicConfig(
    level=logging.DEBUG,
    format="%(asctime)s %(name)-12s %(levelname)-8s %(message)s",
    datefmt="%m-%d %H:%M",
    filename="info.log",
    filemode="w",
)
module_logger = logging.getLogger("exampleExecutable")


class ExampleExecutable(object):
    """
    Simple python executable that resembles the behavior of a real job, i.e., it processes input files, generates
    output files and can perform restarts. Will be mainly used to test the genericJob class.
    """

    def __init__(self):
        print("Execution started")
        self.logger = logging.getLogger("exampleExecutable.module")
        self.logger.info("creating an instance")

        self.input_file = "input.inp"
        self.output_file = "output.log"
        self._count = 0
        self._alat_0 = None
        self._potential = None
        self._alpha = None
        self._alat = None
        self._count = None

    def write_restart(self):
        """
        Write a restart file for a continous simulation divided into multiple jobs.
        """
        with open("restart.out", "w") as f:
            f.write("count {0} \n".format(str(self._count)))

    def read_restart(self):
        """
        Read a restart file for a continous simulation divided into multiple jobs.
        """
        with open("restart.inp", "r") as f:
            line = f.readline()
            _, count = line.split()
            self._count = ast.literal_eval(count)

    def get_energy(self, alat):
        """
        Based on the lattice constant a random energy is calculated.

        Args:
            alat (float): lattice constant

        Returns:
            (list): list of n random energy values, where n equals self._count
        """
        return (
            self._potential(alat - self._alat_0)
            + np.random.random(self._count) * self._alpha
        )

    def run_lib(self, input_dict):
        """
        Run lib executes the job directly in Python instead of calling a separate subprocess, this is faster for all
        Python jobs but not available for non Python jobs. No input or output files are generated when running in
        library mode, instead all input is provided as an input dictionary and the output is returned as a list.

        Args:
            input_dict (dict): input consisting of ["alpha", "alat", "count"]

        Returns:
            list: alat(float), count(int), energy(list)
        """
        n_max = 4  # max. order of polynomial describing the potential
        pot_lst = [float(input_dict["a_" + str(i)]) for i in range(n_max, -1, -1)]
        self._alat_0 = pot_lst[n_max]
        pot_lst[4] = 0
        self._potential = np.poly1d(pot_lst)
        self._alpha = float(input_dict["alpha"])
        self._alat = float(input_dict["alat"])
        self._count = int(input_dict["count"])

        # make the executable a bit more real and throw warnings and error messages
        # depending on the input parameters
        self.logger.debug("type: alpha %s", type(input_dict["alpha"]))
        if self._alpha < 0:
            raise ValueError("noise amplitude alpha < 0")
        if self._count < 1:
            raise ValueError("number of energy steps must be larger than 0")

        if self._alat < 1:
            self.logger.warning("lattice constant alat < 1")

        self.logger.info("Execute program")
        energy = self.get_energy(alat=self._alat)
        return self._alat, self._count, energy

    def run(self):
        """
        Run executes the job as part of a subprocess. The input is written to an input file, then the executable is
        executed and finally the output file is collected and converted to HDF5 format for future processing.
        """
        with open(self.input_file, mode="r") as f:
            input_dict = {}
            for line in f.readlines():
                line = line.strip().split()
                key, value = line[0], line[1]
                # key, value = line.split()
                input_dict[key] = value
                self.logger.info("-> %s %s", key, str(value))

        # parse the input into the correct format
        n_max = 4  # max. order of polynomial describing the potential
        pot_lst = [float(input_dict["a_" + str(i)]) for i in range(n_max, -1, -1)]
        self._alat_0 = pot_lst[n_max]
        pot_lst[4] = 0
        self._potential = np.poly1d(pot_lst)
        self._alpha = float(input_dict["alpha"])
        self._alat = float(input_dict["alat"])
        self._count = int(input_dict["count"])

        # make the executable a bit more real and throw warnings and error messages
        # depending on the input parameters
        self.logger.debug("type: alpha %s", type(input_dict["alpha"]))
        if self._alpha < 0:
            raise ValueError("noise amplitude alpha < 0")
        if self._count < 1:
            raise ValueError("number of energy steps must be larger than 0")

        if self._alat < 1:
            self.logger.warning("lattice constant alat < 1")

        # all values in input_dict are str
        # only the actual code knows which type is needed, i.e., the conversion is done here
        if ast.literal_eval(input_dict["read_restart"]):
            self.logger.info("read restart file")
            self.read_restart()
            self.logger.info("restart file has been successfully read")

        self.logger.info("Execute program")
        energy = self.get_energy(alat=self._alat)

        self.logger.info("Program has been successfully terminated")
        if ast.literal_eval(input_dict["write_restart"]):
            self.logger.info("Write restart file")
            self.write_restart()

        # TODO: use hdf5 output file
        with open(self.output_file, mode="w") as f:
            self.logger.info("Write output file")
            f.write("exampleExecutable logFile \n")
            f.write("alat {0} \n".format(str(self._alat)))
            f.write("count {0} \n".format(str(self._count)))
            for e in energy:
                f.write("energy  {0} \n".format(str(e)))


if __name__ == "__main__":
    ExampleExecutable().run()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from sklearn.cluster import MeanShift
from scipy.sparse import coo_matrix
from pyiron_base import Settings

__author__ = "Joerg Neugebauer, Sam Waseda"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sam Waseda"
__email__ = "waseda@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"

s = Settings()


class Neighbors(object):
    """
    Class for storage of the neighbor information for a given atom based on the KDtree algorithm
    """

    def __init__(self, ref_structure, tolerance=2):
        self.distances = None
        self.vecs = None
        self.indices = None
        self._shells = None
        self._tolerance = tolerance
        self._ref_structure = ref_structure
        self._cluster_vecs = None
        self._cluster_dist = None
        self._boundary_layer_width = 0

    @property
    def shells(self):
        """
            Returns the cell numbers of each atom according to the distances
        """
        return self.get_local_shells(tolerance=self._tolerance)

    def update_vectors(self):
        """
            Update vecs and distances with the same indices
        """
        if np.max(np.absolute(self.vecs)) > 0.49*np.min(np.linalg.norm(self._ref_structure.cell, axis=-1)):
            raise AssertionError('Largest distance value is larger than half the box -> rerun get_neighbors')
        myself = np.ones_like(self.indices)*np.arange(len(self.indices))[:, np.newaxis]
        vecs = self._ref_structure.get_distances(
            myself.flatten(), self.indices.flatten(), mic=np.all(self._ref_structure.pbc), vector=True
        )
        self.vecs = vecs.reshape(self.vecs.shape)
        self.distances = np.linalg.norm(self.vecs, axis=-1)

    def get_local_shells(self, tolerance=None, cluster_by_distances=False, cluster_by_vecs=False):
        """
        Set shell indices based on distances available to each atom. Clustering methods can be used
        at the same time, which will be useful at finite temperature results, but depending on how
        dispersed the atoms are, the algorithm could take some time. If the clustering method(-s)
        have already been launched before this function, it will use the results already available
        and does not execute the clustering method(-s) again.

        Args:
            tolerance (int): decimals in np.round for rounding up distances
            cluster_by_distances (bool): If True, `cluster_by_distances` is called first and the distances obtained
                from the clustered distances are used to calculate the shells. If cluster_by_vecs is True at the same
                time, `cluster_by_distances` will use the clustered vectors for its clustering algorithm. For more,
                see the DocString of `cluster_by_distances`. (default: False)
            cluster_by_vecs (bool): If True, `cluster_by_vectors` is called first and the distances obtained from
                the clustered vectors are used to calculate the shells. (default: False)

        Returns:
            shells (numpy.ndarray): shell indices
        """
        if tolerance is None:
            tolerance = self._tolerance
        if cluster_by_distances:
            if self._cluster_dist is None:
                self.cluster_by_distances(use_vecs=cluster_by_vecs)
            shells = [np.unique(np.round(dist, decimals=tolerance), return_inverse=True)[1]+1
                         for dist in self._cluster_dist.cluster_centers_[self._cluster_dist.labels_].flatten()
                     ]
            return np.array(shells).reshape(self.indices.shape)
        if cluster_by_vecs:
            if self._cluster_vecs is None:
                self.cluster_by_vecs()
            shells = [np.unique(np.round(dist, decimals=tolerance), return_inverse=True)[1]+1
                         for dist in np.linalg.norm(self._cluster_vecs.cluster_centers_[self._cluster_vecs.labels_], axis=-1)
                     ]
            return np.array(shells).reshape(self.indices.shape)
        if self._shells is None:
            if self.distances is None:
                return None
            self._shells = []
            for dist in self.distances:
                self._shells.append(np.unique(np.round(dist[dist<np.inf], decimals=tolerance), return_inverse=True)[1]+1)
            if isinstance(self.distances, np.ndarray):
                self._shells = np.array(self._shells)
        return self._shells

    def get_global_shells(self, tolerance=None, cluster_by_distances=False, cluster_by_vecs=False):
        """
        Set shell indices based on all distances available in the system instead of
        setting them according to the local distances (in contrast to shells defined
        as an attribute in this class). Clustering methods can be used at the same time,
        which will be useful at finite temperature results, but depending on how dispersed
        the atoms are, the algorithm could take some time. If the clustering method(-s)
        have already been launched before this function, it will use the results already
        available and does not execute the clustering method(-s) again.

        Args:
            tolerance (int): decimals in np.round for rounding up distances (default: 2)
            cluster_by_distances (bool): If True, `cluster_by_distances` is called first and the distances obtained
                from the clustered distances are used to calculate the shells. If cluster_by_vecs is True at the same
                time, `cluster_by_distances` will use the clustered vectors for its clustering algorithm. For more,
                see the DocString of `cluster_by_distances`. (default: False)
            cluster_by_vecs (bool): If True, `cluster_by_vectors` is called first and the distances obtained from
                the clustered vectors are used to calculate the shells. (default: False)

        Returns:
            shells (numpy.ndarray): shell indices (cf. shells)
        """
        if tolerance is None:
            tolerance = self._tolerance
        if self.distances is None:
            raise ValueError('neighbors not set')
        distances = self.distances
        if cluster_by_distances:
            if self._cluster_dist is None:
                self.cluster_by_distances(use_vecs=cluster_by_vecs)
            distances = self._cluster_dist.cluster_centers_[self._cluster_dist.labels_].reshape(self.distances.shape)
        elif cluster_by_vecs:
            if self._cluster_vecs is None:
                self.cluster_by_vecs()
            distances = np.linalg.norm(self._cluster_vecs.cluster_centers_[self._cluster_vecs.labels_], axis=-1).reshape(self.distances.shape)
        dist_lst = np.unique(np.round(a=distances, decimals=tolerance))
        shells = distances[:,:,np.newaxis]-dist_lst[np.newaxis,np.newaxis,:]
        shells = np.absolute(shells).argmin(axis=-1)+1
        return shells

    def get_shell_matrix(
        self, chemical_pair=None, cluster_by_distances=False, cluster_by_vecs=False
    ):
        """
        Shell matrices for pairwise interaction. Note: The matrices are always symmetric, meaning if you
        use them as bilinear operators, you have to divide the results by 2.

        Args:
            chemical_pair (list): pair of chemical symbols (e.g. ['Fe', 'Ni'])

        Returns:
            list of sparse matrices for different shells


        Example:
            from pyiron import Project
            structure = Project('.').create_structure('Fe', 'bcc', 2.83).repeat(2)
            J = -0.1 # Ising parameter
            magmoms = 2*np.random.random((len(structure)), 3)-1 # Random magnetic moments between -1 and 1
            neigh = structure.get_neighbors(num_neighbors=8) # Iron first shell
            shell_matrices = neigh.get_shell_matrix()
            print('Energy =', 0.5*J*magmoms.dot(shell_matrices[0].dot(matmoms)))
        """

        pairs = np.stack((self.indices,
            np.ones_like(self.indices)*np.arange(len(self.indices))[:,np.newaxis],
            self.get_global_shells(cluster_by_distances=cluster_by_distances, cluster_by_vecs=cluster_by_vecs)-1),
            axis=-1
        ).reshape(-1, 3)
        shell_max = np.max(pairs[:,-1])+1
        if chemical_pair is not None:
            c = self._ref_structure.get_chemical_symbols()
            pairs = pairs[np.all(np.sort(c[pairs[:,:2]], axis=-1)==np.sort(chemical_pair), axis=-1)]
        shell_matrix = []
        for ind in np.arange(shell_max):
            indices = pairs[ind==pairs[:,-1]]
            if len(indices)>0:
                ind_tmp = np.unique(indices[:,:-1], axis=0, return_counts=True)
                shell_matrix.append(coo_matrix((ind_tmp[1], (ind_tmp[0][:,0], ind_tmp[0][:,1])),
                    shape=(len(self._ref_structure), len(self._ref_structure))
                ))
            else:
                shell_matrix.append(coo_matrix((len(self._ref_structure), len(self._ref_structure))))
        return shell_matrix

    def find_neighbors_by_vector(self, vector, deviation=False):
        """
        Args:
            vector (list/np.ndarray): vector by which positions are translated (and neighbors are searched)
            deviation (bool): whether to return distance between the expect positions and real positions

        Returns:
            np.ndarray: list of id's for the specified translation

        Example:
            a_0 = 2.832
            structure = pr.create_structure('Fe', 'bcc', a_0)
            id_list = structure.find_neighbors_by_vector([0, 0, a_0])
            # In this example, you get a list of neighbor atom id's at z+=a_0 for each atom.
            # This is particularly powerful for SSA when the magnetic structure has to be translated
            # in each direction.
        """

        z = np.zeros(len(self._ref_structure)*3).reshape(-1, 3)
        v = np.append(z[:,np.newaxis,:], self.vecs, axis=1)
        dist = np.linalg.norm(v-np.array(vector), axis=-1)
        indices = np.append(np.arange(len(self._ref_structure))[:,np.newaxis], self.indices, axis=1)
        if deviation:
            return indices[np.arange(len(dist)), np.argmin(dist, axis=-1)], np.min(dist, axis=-1)
        return indices[np.arange(len(dist)), np.argmin(dist, axis=-1)]

    def cluster_by_vecs(self, bandwidth=None, n_jobs=None, max_iter=300):
        """
        Method to group vectors which have similar values. This method should be used as a part of
        neigh.get_global_shells(cluster_by_vecs=True) or neigh.get_local_shells(cluster_by_vecs=True).
        However, in order to specify certain arguments (such as n_jobs or max_iter), it might help to
        have run this function before calling parent functions, as the data obtained with this function
        will be stored in the variable `_cluster_vecs`

        Args:
            bandwidth (float): Resolution (cf. sklearn.cluster.MeanShift)
            n_jobs (int): Number of cores (cf. sklearn.cluster.MeanShift)
            max_iter (int): Number of maximum iterations (cf. sklearn.cluster.MeanShift)
        """
        if bandwidth is None:
            bandwidth = 0.2*np.min(self.distances)
        dr = self.vecs.copy().reshape(-1, 3)
        self._cluster_vecs = MeanShift(bandwidth=bandwidth, n_jobs=n_jobs, max_iter=max_iter).fit(dr)
        self._cluster_vecs.labels_ = self._cluster_vecs.labels_.reshape(self.indices.shape)

    def cluster_by_distances(self, bandwidth=None, use_vecs=False, n_jobs=None, max_iter=300):
        """
        Method to group vectors which have similar values. This method should be used as a part of
        neigh.get_global_shells(cluster_by_vecs=True) or neigh.get_local_shells(cluster_by_distances=True).
        However, in order to specify certain arguments (such as n_jobs or max_iter), it might help to
        have run this function before calling parent functions, as the data obtained with this function
        will be stored in the variable `_cluster_distances`

        Args:
            bandwidth (float): Resolution (cf. sklearn.cluster.MeanShift)
            use_vecs (bool): Whether to form clusters for vecs beforehand. If true, the distances obtained
                from the clustered vectors is used for the distance clustering.  Otherwise neigh.distances
                is used.
            n_jobs (int): Number of cores (cf. sklearn.cluster.MeanShift)
            max_iter (int): Number of maximum iterations (cf. sklearn.cluster.MeanShift)
        """
        if bandwidth is None:
            bandwidth = 0.05*np.min(self.distances)
        dr = self.distances
        if use_vecs:
            if self._cluster_vecs is None:
                self.cluster_by_vecs()
            dr = np.linalg.norm(self._cluster_vecs.cluster_centers_[self._cluster_vecs.labels_], axis=-1)
        self._cluster_dist = MeanShift(bandwidth=bandwidth, n_jobs=n_jobs, max_iter=max_iter).fit(dr.reshape(-1, 1))
        self._cluster_dist.labels_ = self._cluster_dist.labels_.reshape(self.indices.shape)

    def reset_clusters(self, vecs=True, distances=True):
        """
        Method to reset clusters.

        Args:
            vecs (bool): Reset `_cluster_vecs` (cf. `cluster_by_vecs`)
            distances (bool): Reset `_cluster_distances` (cf. `cluster_by_distances`)
        """
        if vecs:
            self._cluster_vecs = None
        if distances:
            self._cluster_distances = None

    def cluster_analysis(
        self, id_list, return_cluster_sizes=False
    ):
        """

        Args:
            id_list:
            return_cluster_sizes:

        Returns:

        """
        self._cluster = [0] * len(self._ref_structure)
        c_count = 1
        # element_list = self.get_atomic_numbers()
        for ia in id_list:
            # el0 = element_list[ia]
            nbrs = self.indices[ia]
            # print ("nbrs: ", ia, nbrs)
            if self._cluster[ia] == 0:
                self._cluster[ia] = c_count
                self.__probe_cluster(c_count, nbrs, id_list)
                c_count += 1

        cluster = np.array(self._cluster)
        cluster_dict = {
            i_c: np.where(cluster == i_c)[0].tolist() for i_c in range(1, c_count)
        }
        if return_cluster_sizes:
            sizes = [self._cluster.count(i_c + 1) for i_c in range(c_count - 1)]
            return cluster_dict, sizes

        return cluster_dict  # sizes

    def __probe_cluster(self, c_count, neighbors, id_list):
        """

        Args:
            c_count:
            neighbors:
            id_list:

        Returns:

        """
        for nbr_id in neighbors:
            if self._cluster[nbr_id] == 0:
                if nbr_id in id_list:  # TODO: check also for ordered structures
                    self._cluster[nbr_id] = c_count
                    nbrs = self.indices[nbr_id]
                    self.__probe_cluster(c_count, nbrs, id_list)

    # TODO: combine with corresponding routine in plot3d
    def get_bonds(self, radius=np.inf, max_shells=None, prec=0.1):
        """

        Args:
            radius:
            max_shells:
            prec: minimum distance between any two clusters (if smaller considered to be single cluster)

        Returns:

        """

        def get_cluster(dist_vec, ind_vec, prec=prec):
            ind_where = np.where(np.diff(dist_vec) > prec)[0] + 1
            ind_vec_cl = [np.sort(group) for group in np.split(ind_vec, ind_where)]
            return ind_vec_cl

        dist = self.distances
        ind = self.indices
        el_list = self._ref_structure.get_chemical_symbols()

        ind_shell = []
        for d, i in zip(dist, ind):
            id_list = get_cluster(d[d < radius], i[d < radius])
            # print ("id: ", d[d<radius], id_list, dist_lst)
            ia_shells_dict = {}
            for i_shell_list in id_list:
                ia_shell_dict = {}
                for i_s in i_shell_list:
                    el = el_list[i_s]
                    if el not in ia_shell_dict:
                        ia_shell_dict[el] = []
                    ia_shell_dict[el].append(i_s)
                for el, ia_lst in ia_shell_dict.items():
                    if el not in ia_shells_dict:
                        ia_shells_dict[el] = []
                    if max_shells is not None:
                        if len(ia_shells_dict[el]) + 1 > max_shells:
                            continue
                    ia_shells_dict[el].append(ia_lst)
            ind_shell.append(ia_shells_dict)
        return ind_shell




# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
import os
from subprocess import Popen, PIPE

from pyiron.vasp.outcar import Outcar
from pyiron.vasp.base import VaspBase
from pyiron.vasp.structure import vasp_sorter
from pyiron.vasp.potential import VaspPotentialSetter
from pyiron.vasp.base import GenericOutput as GenericOutputBase
from pyiron.vasp.base import DFTOutput as DFTOutputBase
from pyiron.vasp.base import Output as OutputBase
from pyiron.atomistics.job.interactive import GenericInteractive


__author__ = "Osamu Waseda, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2018"


class VaspInteractive(VaspBase, GenericInteractive):
    def __init__(self, project, job_name):
        super(VaspInteractive, self).__init__(project, job_name)
        self._interactive_write_input_files = True
        self._interactive_vasprun = None
        self.interactive_flush_frequency = 1

    @property
    def structure(self):
        return GenericInteractive.structure.fget(self)

    @structure.setter
    def structure(self, structure):
        GenericInteractive.structure.fset(self, structure)
        if structure is not None:
            self._potential = VaspPotentialSetter(
                element_lst=structure.get_species_symbols().tolist()
            )

    @property
    def interactive_enforce_structure_reset(self):
        return self._interactive_enforce_structure_reset

    @interactive_enforce_structure_reset.setter
    def interactive_enforce_structure_reset(self, reset):
        raise NotImplementedError(
            "interactive_enforce_structure_reset() is not implemented!"
        )

    def get_structure(self, iteration_step=-1, wrap_atoms=True):
        """
        Gets the structure from a given iteration step of the simulation (MD/ionic relaxation). For static calculations
        there is only one ionic iteration step
        Args:
            iteration_step (int): Step for which the structure is requested
            wrap_atoms (bool): True if the atoms are to be wrapped back into the unit cell

        Returns:
            pyiron.atomistics.structure.atoms.Atoms: The required structure
        """
        if (
            self.server.run_mode.interactive
            or self.server.run_mode.interactive_non_modal
        ):
            structure = GenericInteractive.get_structure(self, iteration_step=iteration_step, wrap_atoms=wrap_atoms)
        else:
            structure = VaspBase.get_structure(self, iteration_step=iteration_step, wrap_atoms=wrap_atoms)

        return structure

    def interactive_close(self):
        if self.interactive_is_activated():
            with open(os.path.join(self.working_directory, "STOPCAR"), "w") as stopcar:
                stopcar.write("LABORT = .TRUE.")  # stopcar.write('LSTOP = .TRUE.')
            try:
                self.run_if_interactive_non_modal()
                self.run_if_interactive_non_modal()
                for atom in self.current_structure.get_scaled_positions():
                    text = " ".join(map("{:19.16f}".format, atom))
                    self._interactive_library.stdin.write(text + "\n")
            except BrokenPipeError:
                self._logger.warning(
                    "VASP calculation exited before interactive_close() - already converged?"
                )
            for key in self.interactive_cache.keys():
                if isinstance(self.interactive_cache[key], list):
                    self.interactive_cache[key] = self.interactive_cache[key]
            super(VaspInteractive, self).interactive_close()
            self.status.collect = True
            self._output_parser = Output()
            if self["vasprun.xml"] is not None:
                self.run()

    def interactive_energy_tot_getter(self):
        return self.interactive_energy_pot_getter()

    def interactive_energy_pot_getter(self):
        if self._interactive_vasprun is not None:
            file_name = os.path.join(self.working_directory, "OUTCAR")
            return self._interactive_vasprun.get_energy_sigma_0(filename=file_name)[-1]
        else:
            return None

    def validate_ready_to_run(self):
        if (
            self.server.run_mode.interactive
            or self.server.run_mode.interactive_non_modal
        ) and "EDIFFG" in self.input.incar._dataset["Parameter"]:
            raise ValueError(
                "If EDIFFG is defined VASP interrupts the interactive run_mode."
            )
        super(VaspInteractive, self).validate_ready_to_run()

    def interactive_forces_getter(self):
        if self._interactive_vasprun is not None:
            file_name = os.path.join(self.working_directory, "OUTCAR")
            forces = self._interactive_vasprun.get_forces(filename=file_name)[-1]
            forces[vasp_sorter(self.structure)] = forces.copy()
            return forces
        else:
            return None

    def interactive_initialize_interface(self):
        if self.executable.executable_path == "":
            self.status.aborted = True
            raise ValueError("No executable set!")
        if self.executable.mpi:
            self._interactive_library = Popen(
                [self.executable.executable_path, str(self.server.cores)],
                stdout=PIPE,
                stdin=PIPE,
                stderr=PIPE,
                cwd=self.working_directory,
                universal_newlines=True,
            )
        else:
            self._interactive_library = Popen(
                self.executable.executable_path,
                stdout=PIPE,
                stdin=PIPE,
                stderr=PIPE,
                cwd=self.working_directory,
                universal_newlines=True,
            )

    def calc_minimize(
        self,
        electronic_steps=400,
        ionic_steps=100,
        max_iter=None,
        pressure=None,
        algorithm=None,
        retain_charge_density=False,
        retain_electrostatic_potential=False,
        ionic_energy=None,
        ionic_forces=None,
        ionic_energy_tolerance=None,
        ionic_force_tolerance=None,
        volume_only=False,
    ):
        """
        Function to setup the hamiltonian to perform ionic relaxations using DFT. The ISIF tag has to be supplied
        separately.

        Args:
            electronic_steps (int): Maximum number of electronic steps
            ionic_steps (int): Maximum number of ionic
            max_iter (int): Maximum number of iterations
            pressure (float): External pressure to be applied
            algorithm (str): Type of VASP algorithm to be used "Fast"/"Accurate"
            retain_charge_density (bool): True if the charge density should be written
            retain_electrostatic_potential (boolean): True if the electrostatic potential should be written
            ionic_energy_tolerance (float): Ionic energy convergence criteria (eV)
            ionic_force_tolerance (float): Ionic forces convergence criteria (overwrites ionic energy) (ev/A)
            ionic_energy (float): Same as ionic_energy_tolerance (deprecated)
            ionic_forces (float): Same as ionic_force_tolerance (deprecated)
            volume_only (bool): Option to relax only the volume (keeping the relative coordinates fixed
        """
        if (
            self.server.run_mode.interactive
            or self.server.run_mode.interactive_non_modal
        ):
            raise NotImplementedError(
                "calc_minimize() is not implemented for the interactive mode use calc_static()!"
            )
        else:
            super(VaspInteractive, self).calc_minimize(
                electronic_steps=electronic_steps,
                ionic_steps=ionic_steps,
                max_iter=max_iter,
                pressure=pressure,
                algorithm=algorithm,
                retain_charge_density=retain_charge_density,
                retain_electrostatic_potential=retain_electrostatic_potential,
                ionic_energy=ionic_energy,
                ionic_forces=ionic_forces,
                volume_only=volume_only,
            )

    def calc_md(
        self,
        temperature=None,
        n_ionic_steps=1000,
        n_print=1,
        time_step=1.0,
        retain_charge_density=False,
        retain_electrostatic_potential=False,
        **kwargs
    ):
        """
        Sets appropriate tags for molecular dynamics in VASP

        Args:
            temperature (int/float/list): Temperature/ range of temperatures in Kelvin
            n_ionic_steps (int): Maximum number of ionic steps
            n_print (int): Prints outputs every n_print steps
            time_step (float): time step (fs)
            retain_charge_density (bool): True id the charge density should be written
            retain_electrostatic_potential (bool): True if the electrostatic potential should be written
        """
        if (
            self.server.run_mode.interactive
            or self.server.run_mode.interactive_non_modal
        ):
            raise NotImplementedError(
                "calc_md() is not implemented for the interactive mode use calc_static()!"
            )
        else:
            super(VaspInteractive, self).calc_md(
                temperature=temperature,
                n_ionic_steps=n_ionic_steps,
                time_step=time_step,
                n_print=n_print,
                retain_charge_density=retain_charge_density,
                retain_electrostatic_potential=retain_electrostatic_potential,
                **kwargs
            )

    def run_if_interactive_non_modal(self):
        initial_run = self.interactive_is_activated()
        super(VaspInteractive, self).run_if_interactive()
        if initial_run:
            atom_numbers = self.current_structure.get_number_species_atoms()
            for species in atom_numbers.keys():
                indices = self.current_structure.select_index(species)
                for i in indices:
                    text = " ".join(
                        map(
                            "{:19.16f}".format,
                            self.current_structure.get_scaled_positions()[i],
                        )
                    )
                    self._logger.debug("Vasp library: " + text)
                    self._interactive_library.stdin.write(text + "\n")
            self._interactive_library.stdin.flush()
        self._interactive_fetch_completed = False

    def run_if_interactive(self):
        self.run_if_interactive_non_modal()
        self._interactive_check_output()
        self._interactive_vasprun = Outcar()
        self.interactive_collect()

    def interactive_fetch(self):
        if (
            self._interactive_fetch_completed
            and self.server.run_mode.interactive_non_modal
        ):
            print("First run and then fetch")
        else:
            self._interactive_check_output()
            self._interactive_vasprun = Outcar()
            super(VaspInteractive, self).interactive_collect()
            self._logger.debug("interactive run - done")

    def interactive_positions_setter(self, positions):
        pass

    def _check_incar_parameter(self, parameter, value):
        if parameter not in self.input.incar._dataset["Parameter"]:
            self.input.incar[parameter] = value

    def _interactive_check_output(self):
        while self._interactive_library.poll() is None:
            text = self._interactive_library.stdout.readline()
            if "POSITIONS: reading from stdin" in text:
                return

    def _run_if_new(self, debug=False):
        if (
            self.server.run_mode.interactive
            or self.server.run_mode.interactive_non_modal
        ):
            self._check_incar_parameter(parameter="INTERACTIVE", value=True)
            self._check_incar_parameter(parameter="IBRION", value=-1)
            self._check_incar_parameter(parameter="POTIM", value=0.0)
            self._check_incar_parameter(parameter="NSW", value=1000)
            self._check_incar_parameter(parameter="ISYM", value=0)
        super(VaspInteractive, self)._run_if_new(debug=debug)


class Output(OutputBase):
    """
    Handles the output from a VASP simulation.

    Attributes:
        electronic_structure: Gives the electronic structure of the system
        electrostatic_potential: Gives the electrostatic/local potential of the system
        charge_density: Gives the charge density of the system
    """

    def __init__(self):
        super(Output, self).__init__()
        self.generic_output = GenericOutput()
        self.dft_output = DFTOutput()


class GenericOutput(GenericOutputBase):
    """

    This class stores the generic output like different structures, energies and forces from a simulation in a highly
    generic format. Usually the user does not have to access this class.

    Attributes:
        log_dict (dict): A dictionary of all tags and values of generic data (positions, forces, etc)
    """

    def __init__(self):
        super(GenericOutput, self).__init__()

    def to_hdf(self, hdf):
        """
        Save the object in a HDF5 file

        Args:
            hdf (pyiron_base.generic.hdfio.ProjectHDFio): HDF path to which the object is to be saved

        """
        with hdf.open("generic") as hdf_go:
            # hdf_go["description"] = self.description
            for key, val in self.log_dict.items():
                hdf_go[key] = val
            with hdf_go.open("dft") as hdf_dft:
                for key, val in self.dft_log_dict.items():
                    hdf_dft[key] = val
                if self.bands.eigenvalue_matrix is not None:
                    self.bands.to_hdf(hdf_dft, "bands")


class DFTOutput(DFTOutputBase):
    """
    This class stores the DFT specific output

    Attributes:
        log_dict (dict): A dictionary of all tags and values of DFT data
    """

    def __init__(self):
        super(DFTOutput, self).__init__()

    def to_hdf(self, hdf):
        """
        Save the object in a HDF5 file

        Args:
            hdf (pyiron_base.generic.hdfio.ProjectHDFio): HDF path to which the object is to be saved

        """
        with hdf.open("dft") as hdf_dft:
            for key, val in self.log_dict.items():
                hdf_dft[key] = val



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import codecs
from collections import OrderedDict
from datetime import datetime
import dill as pickle
import inspect
import json
import numpy as np
import os
import pandas
from pandas.errors import EmptyDataError
from tqdm import tqdm
import types

from pyiron_base import GenericJob, FileHDFio, get_function_from_string
from pyiron.table.funct import (
    get_incar,
    get_sigma,
    get_total_number_of_atoms,
    get_elements,
    get_convergence_check,
    get_number_of_species,
    get_number_of_ionic_steps,
    get_ismear,
    get_encut,
    get_n_kpts,
    get_n_equ_kpts,
    get_number_of_final_electronic_steps,
    get_majority_species,
    get_job_name,
    get_job_id,
    get_energy_tot,
    get_energy_pot,
    get_energy_free,
    get_energy_int,
    get_energy_tot_per_atom,
    get_energy_pot_per_atom,
    get_energy_free_per_atom,
    get_energy_int_per_atom,
    get_e_conv_level,
    get_f_states,
    get_e_band,
    get_majority_crystal_structure,
    get_equilibrium_parameters,
    get_structure,
    get_forces,
    get_magnetic_structure,
    get_average_waves,
    get_plane_waves,
    get_ekin_error,
    get_volume,
    get_volume_per_atom,
)


__author__ = "Uday Gajera, Jan Janssen, Joerg Neugebauer"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "0.0.1"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class FunctionContainer(object):
    """
    Class which is able to append, store and retreive a set of functions.

    """
    def __init__(self):
        self._user_function_dict = {}
        self._system_function_lst = [
            get_incar,
            get_sigma,
            get_total_number_of_atoms,
            get_elements,
            get_convergence_check,
            get_number_of_species,
            get_number_of_ionic_steps,
            get_ismear,
            get_encut,
            get_n_kpts,
            get_n_equ_kpts,
            get_number_of_final_electronic_steps,
            get_majority_species,
            get_job_name,
            get_job_id,
            get_energy_tot,
            get_energy_pot,
            get_energy_free,
            get_energy_int,
            get_energy_tot_per_atom,
            get_energy_pot_per_atom,
            get_energy_free_per_atom,
            get_energy_int_per_atom,
            get_e_conv_level,
            get_f_states,
            get_e_band,
            get_majority_crystal_structure,
            get_equilibrium_parameters,
            get_structure,
            get_forces,
            get_magnetic_structure,
            get_average_waves,
            get_plane_waves,
            get_ekin_error,
            get_volume,
            get_volume_per_atom,
        ]
        self._system_function_dict = {
            func.__name__: False for func in self._system_function_lst
        }
        self._system_function_dict["get_job_id"] = True

    @property
    def _function_lst(self):
        return [
            funct
            for funct in self._system_function_lst
            if funct.__name__ in self._system_function_dict.keys()
            and self._system_function_dict[funct.__name__]
        ] + list(self._user_function_dict.values())

    def _to_hdf(self, hdf):
        self._to_pickle(
            hdf=hdf, key="user_function_dict", value=self._user_function_dict
        )
        self._to_pickle(
            hdf=hdf, key="system_function_dict", value=self._system_function_dict
        )

    def _from_hdf(self, hdf):
        self._user_function_dict = self._from_pickle(hdf=hdf, key="user_function_dict")
        self._system_function_dict = self._from_pickle(
            hdf=hdf, key="system_function_dict"
        )

    def __setitem__(self, key, item):
        if isinstance(item, str):
            self._user_function_dict[key] = eval(
                'lambda job: {"' + key + '":' + item + "}"
            )
        elif isinstance(item, types.FunctionType):
            self._user_function_dict[key] = lambda job: {key: item(job)}
        else:
            raise TypeError("unsupported function type!")

    def __getitem__(self, key):
        return self._user_function_dict[key]

    def __getattr__(self, name):
        if name in list(self._system_function_dict.keys()):
            self._system_function_dict[name] = True
            return self._system_function_dict[name]
        else:
            super(FunctionContainer, self).__getattr__(name)

    def __dir__(self):
        return list(self._system_function_dict.keys())

    @staticmethod
    def _to_pickle(hdf, key, value):
        hdf[key] = codecs.encode(pickle.dumps(value), "base64").decode()

    @staticmethod
    def _from_pickle(hdf, key):
        return pickle.loads(codecs.decode(hdf[key].encode(), "base64"))


class JobFilters(object):
    """
    Certain predefined job filters

    """
    @staticmethod
    def job_type(job_type):
        def filter_job_type(job):
            return job.__name__ == job_type
        return filter_job_type

    @staticmethod
    def job_name_contains(job_name_segment):
        def filter_job_name_segment(job):
            return job_name_segment in job.job_name
        return filter_job_name_segment


class PyironTable(object):
    """
    Class for easy, efficient, and pythonic analysis of data from pyiron projects

    Args:
        project (pyiron.project.Project/None): The project to analyze
        name (str): Name of the pyiron table
    """
    def __init__(self, project, name=None):
        self._project = project
        self._df = pandas.DataFrame({})
        self.convert_to_object = False
        self._name = name
        self._db_filter_function = always_true_pandas
        self._db_filter_function_str = inspect.getsource(always_true_pandas)
        self._filter_function = always_true
        self._filter_function_str = inspect.getsource(always_true)
        self._filter = JobFilters()
        self.add = FunctionContainer()
        self._csv_file = None
        if self._is_file():
            self.load()

        self.EMPTY_STR = "-"

    @property
    def filter(self):
        """
        Object containing pre-defined filter functions

        Returns:
            pyiron.table.datamining.JobFilters: The object containing the filters

        """
        return self._filter

    @property
    def _file_name_csv(self):
        if self._csv_file is None:
            return self._project.path + self.name + ".csv"
        else:
            return self._csv_file

    @property
    def _file_name_txt(self):
        return self._project.path + self.name + ".txt"

    @property
    def name(self):
        """
        Name of the table. Takes the project name if not specified

        Returns:
            str: Name of the table

        """
        if self._name is None:
            return self._project.name
        return self._name

    @property
    def db_filter_function(self):
        """
        Function to filter the a project database table before job specific functions are applied.

        The function must take a pyiron project table in the pandas.DataFrame format (project.job_table()) and return a
        boolean pandas.DataSeries with the same number of rows as the project table

        Example:

            def function(df):
                return (df["chemicalformula"=="H2"]) & (df["hamilton"=="Vasp"])

        """
        return self._db_filter_function

    @db_filter_function.setter
    def db_filter_function(self, funct):
        self._db_filter_function = funct
        try:
            self._db_filter_function_str = inspect.getsource(funct)
        except (OSError, IOError):
            pass

    @property
    def filter_function(self):
        """
        Function to filter each job before more expensive functions are applied
        """
        return self._filter_function

    @filter_function.setter
    def filter_function(self, funct):
        self._filter_function = funct
        try:
            self._filter_function_str = inspect.getsource(funct)
        except (OSError, IOError):
            pass

    def to_hdf(self):
        file = FileHDFio(file_name=self._project.path + self.name + ".h5", h5_path="/")
        self.add._to_hdf(file)

    def from_hdf(self):
        file = FileHDFio(file_name=self._project.path + self.name + ".h5", h5_path="/")
        self.add._from_hdf(file)

    def save(self, name=None):
        self._name = name
        self.to_hdf()
        self._save_csv()

    def load(self, name=None):
        self._name = name
        self.from_hdf()
        self._load_csv()

    def create_table(self, enforce_update=False, level=3, file=None, job_status_list=None):
        skip_table_update = False
        filter_funct = self.filter_function
        if job_status_list is None:
            job_status_list = ["finished"]
        if self._is_file():
            if file is None:
                file = FileHDFio(
                    file_name=self._project.path + self.name + ".h5", h5_path="/"
                )
            temp_user_function_dict, temp_system_function_dict = self._get_data_from_hdf5(
                hdf=file
            )
            job_update_lst = self._collect_job_update_lst(
                job_status_list=job_status_list,
                filter_funct=filter_funct,
                job_stored_ids=self._get_job_ids()
            )
            keys_update_user_lst = [
                key
                for key in self.add._user_function_dict.keys()
                if key not in temp_user_function_dict.keys()
            ]
            keys_update_system_lst = [
                k
                for k, v in self.add._system_function_dict.items()
                if v and not temp_system_function_dict[k]
            ]
            if (
                len(job_update_lst) == 0
                and len(keys_update_user_lst) == 0
                and keys_update_system_lst == 0
                and not enforce_update
            ):
                skip_table_update = True
        else:
            job_update_lst = self._collect_job_update_lst(
                job_status_list=job_status_list,
                filter_funct=filter_funct,
                job_stored_ids=None
            )
            keys_update_user_lst, keys_update_system_lst = [], []
        if not skip_table_update and len(job_update_lst) != 0:
            df_new_ids = self._iterate_over_job_lst(
                job_lst=job_update_lst, function_lst=self.add._function_lst, level=level
            )
        else:
            df_new_ids = pandas.DataFrame({})
        if not skip_table_update and (
            len(keys_update_user_lst) != 0 or len(keys_update_system_lst) != 0
        ):
            job_update_lst = self._collect_job_update_lst(
                job_status_list=job_status_list,
                filter_funct=filter_funct,
                job_stored_ids=None
            )
            function_lst = [
                v
                for k, v in self.add._user_function_dict.items()
                if k in keys_update_system_lst
            ] + [
                funct
                for funct in self.add._system_function_lst
                if funct.__name__ in keys_update_system_lst
            ]
            df_new_keys = self._iterate_over_job_lst(
                job_lst=job_update_lst, function_lst=function_lst, level=level
            )
        else:
            df_new_keys = pandas.DataFrame({})
        if len(self._df) > 0 and len(df_new_keys) > 0:
            self._df = pandas.concat(
                [self._df, df_new_keys], axis=1, sort=False
            ).reset_index(drop=True)
        if len(self._df) > 0 and len(df_new_ids) > 0:
            self._df = pandas.concat([self._df, df_new_ids], sort=False).reset_index(
                drop=True
            )
        elif len(df_new_ids) > 0:
            self._df = df_new_ids

    def convert_dict(self, input_dict):
        return {key: self.str_to_value(value) for key, value in input_dict.items()}

    def refill_dict(self, diff_dict_lst):
        total_key_lst = self.total_lst_of_keys(diff_dict_lst)
        for ind, sub_dict in enumerate(diff_dict_lst):
            for key in total_key_lst:
                if key not in sub_dict.keys():
                    sub_dict[key] = self.EMPTY_STR
                else:
                    sub_dict[key] = self.str_to_value(sub_dict[key])

    def col_to_value(self, col_name):
        val_lst, key_lst, ind_lst = [], [], []
        for ind, name in enumerate(self._df[col_name]):
            #             print ('name: ', ind, name)
            #             if name == self.EMPTY_STR:
            #                 continue
            name = name.split("_")
            ind_lst.append(ind)
            key_lst.append(name[0])
            val_lst.append(eval(".".join(name[1:])))
        if len(set(key_lst)) == 1:
            key = key_lst[0]
            self._df[key] = val_lst
        else:
            raise ValueError("key not unique: {}".format(set(key_lst)))

    def get_dataframe(self):
        return self._df

    def list_nodes(self):
        return list(self._df.columns)

    def list_groups(self):
        return list(set(self._df["col_0"]))

    @staticmethod
    def str_to_value(input_val):
        if not isinstance(input_val, str):
            return input_val
        else:
            try:
                return eval(input_val)
            except (TypeError, SyntaxError, NameError):
                return input_val

    @staticmethod
    def _apply_function_on_job(funct, job):
        try:
            return funct(job)
        except ValueError:
            return {}

    @staticmethod
    def total_lst_of_keys(diff_dict_lst):
        total_key_lst = []
        for sub_dict in diff_dict_lst:
            for key in sub_dict.keys():
                total_key_lst.append(key)
        return set(total_key_lst)

    def __getitem__(self, item, max_level=5):
        rename_dict = OrderedDict()
        if item in self.list_groups():
            for i in range(1, max_level):
                rename_dict["col_{}".format(i)] = "col_{}".format(i - 1)

            new_table = PyironTable(self._project[item])
            new_table._df = self._df.drop("col_0", axis=1)
            new_table._df.rename(index=str, columns=rename_dict, inplace=True)
            return new_table
        if item in self.list_nodes():
            return np.array(self._df[item])
        return None

    def __str__(self):
        return self._df.__str__()

    def __repr__(self):
        """
        Human readable string representation

        Returns:
            str: pandas Dataframe structure as string
        """
        return self._df.__repr__()

    def _is_file(self):
        return self._project is not None and os.path.isfile(self._file_name_csv)

    def _save_csv(self):
        self._df.to_csv(self._file_name_csv, index=False)

    def _load_csv(self):
        self._df = pandas.read_csv(self._file_name_csv)

    def _get_project_list(self, name, pr_len, level=3):
        lst = [self.EMPTY_STR for _ in range(level)]
        for i, p in enumerate(name.split("/")[pr_len - 1 : -1]):
            if len(lst) > i:
                lst[i] = p
        return lst

    def _get_data_from_hdf5(self, hdf):
        temp_user_function_dict = self.add._from_pickle(
            hdf=hdf, key="user_function_dict"
        )
        temp_system_function_dict = self.add._from_pickle(
            hdf=hdf, key="system_function_dict"
        )
        return temp_user_function_dict, temp_system_function_dict

    def _get_job_ids(self):
        if len(self._df) > 0:
            return self._df.job_id.values
        else:
            return np.array([])

    def _get_filtered_job_ids_from_project(self, recursive=True):
        project_table = self._project.job_table(recursive=recursive)
        filter_funct = self.db_filter_function
        return project_table[filter_funct(project_table)]["id"].tolist()

    def _apply_list_of_functions_on_job(self, job, fucntion_lst):
        diff_dict = {}
        for funct in fucntion_lst:
            funct_dict = self._apply_function_on_job(funct, job)
            for key, value in funct_dict.items():
                diff_dict[key] = value
        return diff_dict

    def _iterate_over_job_lst(self, job_lst, function_lst, level):
        pr_len = len(self._project.project_path.split("/"))
        diff_dict_lst = []
        for job_inspect in tqdm(job_lst):
            if self.convert_to_object:
                job = job_inspect.load_object()
            else:
                job = job_inspect
            diff_dict = self._apply_list_of_functions_on_job(
                job=job, fucntion_lst=function_lst
            )
            pr_lst = self._get_project_list(job.project.project_path, pr_len, level)
            for ic, col in enumerate(pr_lst):
                diff_dict["col_{}".format(ic)] = col
            diff_dict_lst.append(diff_dict)
        self.refill_dict(diff_dict_lst)
        return pandas.DataFrame(diff_dict_lst)

    def _collect_job_update_lst(self, job_status_list, filter_funct, job_stored_ids=None):
        """
        Collect jobs to update the pyiron table

        Args:
            job_status_list (list): List of job status to consider
            filter_funct (function): Filter function
            job_stored_ids (list/ None): List of already analysed job ids

        Returns:
            list: List of JobCore objects
        """
        if job_stored_ids is not None:
            job_id_lst = [
                job_id
                for job_id in self._get_filtered_job_ids_from_project()
                if job_id not in job_stored_ids
            ]
        else:
            job_id_lst = self._get_filtered_job_ids_from_project()

        job_update_lst = []
        for job_id in tqdm(job_id_lst):
            try:
                job = self._project.inspect(job_id)
            except IndexError:  # In case the job was deleted while the pyiron table is running
                job = None
            if job is not None and job.status in job_status_list and filter_funct(job):
                job_update_lst.append(job)
        return job_update_lst

    def _repr_html_(self):
        """
        Internal helper function to represent the GenericParameters object within the Jupyter Framework

        Returns:
            HTML: Jupyter HTML object
        """
        return self._df._repr_html_()


class TableJob(GenericJob):
    def __init__(self, project, job_name):
        super(TableJob, self).__init__(project, job_name)
        self.__version__ = "0.1"
        self.__name__ = "TableJob"
        self._analysis_project = None
        self._pyiron_table = PyironTable(project=None)
        self._enforce_update = False
        self._project_level = 0
        self.analysis_project = project.project

    @property
    def filter(self):
        return self._pyiron_table.filter

    @property
    def project_level(self):
        return self._project_level

    @project_level.setter
    def project_level(self, level):
        self._project_level = level

    @property
    def db_filter_function(self):
        return self._pyiron_table.db_filter_function

    @db_filter_function.setter
    def db_filter_function(self, funct):
        self._pyiron_table.db_filter_function = funct

    @property
    def filter_function(self):
        return self._pyiron_table.filter_function

    @filter_function.setter
    def filter_function(self, funct):
        self._pyiron_table.filter_function = funct

    @property
    def pyiron_table(self):
        return self._pyiron_table

    @property
    def ref_project(self):
        return self.analysis_project

    @ref_project.setter
    def ref_project(self, project):
        self.analysis_project = project

    @property
    def analysis_project(self):
        return self._analysis_project

    @analysis_project.setter
    def analysis_project(self, project):
        self._analysis_project = project
        self._pyiron_table = PyironTable(project=self._analysis_project)

    @property
    def add(self):
        return self._pyiron_table.add

    @property
    def convert_to_object(self):
        return self._pyiron_table.convert_to_object

    @convert_to_object.setter
    def convert_to_object(self, conv_to_obj):
        self._pyiron_table.convert_to_object = conv_to_obj

    @property
    def enforce_update(self):
        return self._enforce_update

    @enforce_update.setter
    def enforce_update(self, enforce):
        if isinstance(enforce, bool):
            if enforce:
                self._enforce_update = True
                if self.status.finished:
                    self.status.created = True
            else:
                self._enforce_update = False
        else:
            raise TypeError()

    @staticmethod
    def convert_numpy_to_list(table_dict):
        for k, v in table_dict.items():
            for k1, v1 in v.items():
                if isinstance(v1, np.ndarray):
                    v[k1] = v1.tolist()
        return table_dict

    def to_hdf(self, hdf=None, group_name=None):
        """
        Store pyiron table job in HDF5

        Args:
            hdf:
            group_name:

        """
        super(TableJob, self).to_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            hdf5_input["bool_dict"] = {
                "enforce_update": self._enforce_update,
                "convert_to_object": self._pyiron_table.convert_to_object,
            }
            self._pyiron_table.add._to_hdf(hdf5_input)
            if self._analysis_project is not None:
                hdf5_input["project"] = {
                    "path": self._analysis_project.path,
                    "user": self._analysis_project.user,
                    "sql_query": self._analysis_project.sql_query,
                    "filter": self._analysis_project._filter,
                    "inspect_mode": self._analysis_project._inspect_mode,
                }
            if self.pyiron_table._filter_function is not None:
                try:
                    hdf5_input["filter"] = inspect.getsource(
                        self.pyiron_table._filter_function
                    )
                except (OSError, IOError):
                    if self.pyiron_table._filter_function_str is not None:
                        hdf5_input["filter"] = self.pyiron_table._filter_function_str
            if self.pyiron_table._db_filter_function is not None:
                try:
                    hdf5_input["db_filter"] = inspect.getsource(
                        self.pyiron_table._db_filter_function
                    )
                except (OSError, IOError):
                    if self.pyiron_table._db_filter_function_str is not None:
                        hdf5_input["db_filter"] = self.pyiron_table._db_filter_function_str
        if len(self.pyiron_table._df) != 0:
            with self.project_hdf5.open("output") as hdf5_output:
                table_dict = self.convert_numpy_to_list(self.pyiron_table._df.to_dict())
                hdf5_output["table"] = json.dumps(table_dict)

    def from_hdf(self, hdf=None, group_name=None):
        """
        Restore pyiron table job from HDF5

        Args:
            hdf:
            group_name:
        """
        super(TableJob, self).from_hdf(hdf=hdf, group_name=group_name)
        with self.project_hdf5.open("input") as hdf5_input:
            if "project" in hdf5_input.list_nodes():
                project_dict = hdf5_input["project"]
                project = self.project.__class__(
                    path=project_dict["path"],
                    user=project_dict["user"],
                    sql_query=project_dict["sql_query"],
                )
                project._filter = project_dict["filter"]
                project._inspect_mode = project_dict["inspect_mode"]
                self.analysis_project = project
            if "filter" in hdf5_input.list_nodes():
                self.pyiron_table._filter_function_str = hdf5_input["filter"]
                self.pyiron_table.filter_function = get_function_from_string(
                    hdf5_input["filter"]
                )
            if "db_filter" in hdf5_input.list_nodes():
                self.pyiron_table._db_filter_function_str = hdf5_input["db_filter"]
                self.pyiron_table.db_filter_function = get_function_from_string(
                    hdf5_input["db_filter"]
                )
            bool_dict = hdf5_input["bool_dict"]
            self._enforce_update = bool_dict["enforce_update"]
            self._pyiron_table.convert_to_object = bool_dict["convert_to_object"]
            self._pyiron_table.add._from_hdf(hdf5_input)
        pyiron_table = os.path.join(self.working_directory, "pyirontable.csv")
        if os.path.exists(pyiron_table):
            try:
                self._pyiron_table._df = pandas.read_csv(pyiron_table)
                self._pyiron_table._csv_file = pyiron_table
            except EmptyDataError:
                pass
        else:
            with self.project_hdf5.open("output") as hdf5_output:
                if "table" in hdf5_output.list_nodes():
                    self._pyiron_table._df = pandas.DataFrame(
                        json.loads(hdf5_output["table"])
                    )

    def validate_ready_to_run(self):
        if self._analysis_project is None:
            raise ValueError("Analysis project not defined!")

    def run_static(self):
        self._create_working_directory()
        self.status.running = True
        self.update_table()
        self.status.finished = True

    def update_table(self, job_status_list=None):
        """
        Update the pyiron table object, add new columns if a new function was added or add new rows for new jobs

        Args:
            job_status_list (list/None): List of job status which are added to the table by default ["finished"]
        """
        if job_status_list is None:
            job_status_list = ["finished"]
        self.project.db.item_update({"timestart": datetime.now()}, self.job_id)
        with self.project_hdf5.open("input") as hdf5_input:
            self._pyiron_table.create_table(
                enforce_update=self._enforce_update,
                file=hdf5_input,
                level=self._project_level,
                job_status_list=job_status_list,
            )
        self.to_hdf()
        self._pyiron_table._df.to_csv(
            os.path.join(self.working_directory, "pyirontable.csv"), index=False
        )
        with self.project_hdf5.open("output") as hdf5_output:
            table_dict = self.convert_numpy_to_list(self.pyiron_table._df.to_dict())
            hdf5_output["table"] = json.dumps(table_dict)
        self.project.db.item_update(self._runtime(), self.job_id)

    def write_input(self):
        pass

    def get_dataframe(self):
        """

        Returns:
            pandas.Dataframe
        """
        return self.pyiron_table._df


def always_true_pandas(job_table):
    """
    A function which returns a pandas Series with all True values based on the size of the input pandas dataframe
    Args:
        job_table (pandas.DataFrame): Input dataframe

    Returns:
        pandas.Series: A series of True values

    """
    from pandas import Series
    return Series([True] * len(job_table), index=job_table.index)


def always_true(_):
    """
    A function that always returns True no matter what!

    Returns:
        bool: True

    """
    return True



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import sys

try:
    from ase.atoms import Atoms as ASEAtoms
except ImportError:
    pass

__author__ = "Joerg Neugebauer"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


def publication():
    return {
        "ase": {
            "ase-paper": {
                "author": [
                    "Ask Hjorth Larsen",
                    "Jens Jørgen Mortensen",
                    "Jakob Blomqvist",
                    "Ivano E Castelli",
                    "Rune Christensen",
                    "Marcin Dułak",
                    "Jesper Friis",
                    "Michael N Groves",
                    "Bjørk Hammer",
                    "Cory Hargus",
                    "Eric D Hermes",
                    "Paul C Jennings",
                    "Peter Bjerre Jensen",
                    "James Kermode",
                    "John R Kitchin",
                    "Esben Leonhard Kolsbjerg",
                    "Joseph Kubal",
                    "Kristen Kaasbjerg",
                    "Steen Lysgaard",
                    "Jón Bergmann Maronsson",
                    "Tristan Maxson",
                    "Thomas Olsen",
                    "Lars Pastewka",
                    "Andrew Peterson",
                    "Carsten Rostgaard",
                    "Jakob Schiøtz",
                    "Ole Schütt",
                    "Mikkel Strange",
                    "Kristian S Thygesen",
                    "Tejs Vegge",
                    "Lasse Vilhelmsen",
                    "Michael Walter",
                    "Zhenhua Zeng",
                    "Karsten W Jacobsen",
                ],
                "title": "The atomic simulation environment—a Python library for working with atoms",
                "journal": "Journal of Physics: Condensed Matter",
                "volume": "29",
                "number": "27",
                "pages": "273002",
                "url": "http://stacks.iop.org/0953-8984/29/i=27/a=273002",
                "year": "2017",
            }
        }
    }



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from collections import OrderedDict
import hashlib
import numpy as np
import warnings
from pyiron_base import GenericParameters
import scipy.constants as spc

__author__ = "Joerg Neugebauer, Sudarsan Surendralal, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Sep 1, 2017"


# Conversion factors for transfroming pyiron units to Lammps units (alphabetical)
AMU_TO_G = spc.atomic_mass * spc.kilo
AMU_TO_KG = spc.atomic_mass
ANG_PER_FS_TO_ANG_PER_PS = spc.pico / spc.femto
ANG_PER_FS_TO_BOHR_PER_FS = spc.angstrom / spc.physical_constants['Bohr radius'][0]
ANG_PER_FS_TO_CM_PER_S = (spc.angstrom / spc.femto) / spc.centi
ANG_PER_FS_TO_M_PER_S = spc.angstrom / spc.femto
ANG_TO_BOHR = spc.angstrom / spc.physical_constants['Bohr radius'][0]
ANG_TO_CM = spc.angstrom / spc.centi
ANG_TO_M = spc.angstrom
EL_TO_COUL = spc.elementary_charge
EV_PER_ANG_TO_DYNE = (spc.electron_volt / spc.angstrom) / spc.dyne
EV_PER_ANG_TO_HA_PER_BOHR = spc.physical_constants["electron volt-hartree relationship"][0] * \
                            spc.physical_constants['Bohr radius'][0] / spc.angstrom
EV_PER_ANG_TO_KCAL_PER_MOL_ANG = spc.eV / (spc.kilo * spc.calorie / spc.N_A)
EV_PER_ANG_TO_N = spc.electron_volt / spc.angstrom
EV_TO_ERG = spc.electron_volt / spc.erg
EV_TO_HA = spc.physical_constants["electron volt-hartree relationship"][0]
EV_TO_J = spc.electron_volt
EV_TO_KCAL_PER_MOL = spc.eV / (spc.kilo * spc.calorie / spc.N_A)
FS_TO_PS = spc.femto / spc.pico
FS_TO_S = spc.femto
GPA_TO_ATM = spc.giga / spc.atm
GPA_TO_BAR = spc.giga / spc.bar
GPA_TO_BARYE = spc.giga / (spc.micro * spc.bar)  # "barye" = 1e-6 bar
GPA_TO_PA = spc.giga

# Conversions for most of the Lammps units to Pyiron units
# Lammps units source doc: https://lammps.sandia.gov/doc/units.html
# Pyrion units source doc: https://pyiron.readthedocs.io/en/latest/source/faq.html
# At time of writing, not all these conversion factors are used, but may be helpful later.
LAMMPS_UNIT_CONVERSIONS = {
    "metal": {
        "mass": 1.,
        "distance": 1.,
        "time": FS_TO_PS,
        "energy": 1.,
        "velocity": ANG_PER_FS_TO_ANG_PER_PS,
        "force": 1.,
        "temperature": 1.,
        "pressure": GPA_TO_BAR,
        "charge": 1.
    },
    "si": {
        "mass": AMU_TO_KG,
        "distance": ANG_TO_M,
        "time": FS_TO_S,
        "energy": EV_TO_J,
        "velocity": ANG_PER_FS_TO_M_PER_S,
        "force": EV_PER_ANG_TO_N,
        "temperature": 1.,
        "pressure": GPA_TO_PA,
        "charge": EL_TO_COUL
    },
    "cgs": {
        "mass": AMU_TO_G,
        "distance": ANG_TO_CM,
        "time": FS_TO_S,
        "energy": EV_TO_ERG,
        "velocity": ANG_PER_FS_TO_CM_PER_S,
        "force": EV_PER_ANG_TO_DYNE,
        "temperature": 1.,
        "pressure": GPA_TO_BARYE,
        "charge": 4.8032044e-10  # In statCoulombs, but these are deprecated and thus not in scipt.constants
    },
    "real": {
        "mass": 1.,
        "distance": 1.,
        "time": 1.,
        "energy": EV_TO_KCAL_PER_MOL,
        "velocity": 1.,
        "force": EV_PER_ANG_TO_KCAL_PER_MOL_ANG,
        "temperature": 1.,
        "pressure": GPA_TO_ATM,
        "charge": 1.
    },
    "electron": {
        "mass": 1.,
        "distance": ANG_TO_BOHR,
        "time": 1.,
        "energy": EV_TO_HA,
        "velocity": ANG_PER_FS_TO_BOHR_PER_FS,
        "force": EV_PER_ANG_TO_HA_PER_BOHR,
        "temperature": 1.,
        "pressure": GPA_TO_PA,
        "charge": 1.
    },
}


class LammpsControl(GenericParameters):
    def __init__(self, input_file_name=None, **qwargs):
        super(LammpsControl, self).__init__(
            input_file_name=input_file_name, table_name="control_inp", comment_char="#"
        )
        self._init_block_dict()

    @property
    def dataset(self):
        return self._dataset

    def _init_block_dict(self):
        block_dict = OrderedDict()
        block_dict["read_restart"] = "read_restart"
        block_dict["structure"] = (
            "units",
            "dimension",
            "boundary",
            "atom_style",
            "read_data",
            "include",
        )
        block_dict["potential"] = (
            "group",
            "set",
            "pair_style",
            "pair_coeff",
            "bond_style",
            "bond_coeff",
            "angle_style",
            "angle_coeff",
            "kspace_style",
            "neighbor",
        )
        block_dict["compute"] = (
            "compute",
            "fix",
            "variable",
            "fix_modify",
        )
        block_dict["setup"] = (
            "min_style",
            "min_modify",
            "neigh_modify",
            "velocity",
            "timestep",
            "dielectric",
            "reset_timestep",
        )
        block_dict["control"] = (
            "dump",
            "dump_modify",
            "thermo_style",
            "thermo_modify",
            "thermo",
        )
        block_dict["run"] = ("run", "minimize")
        block_dict["write_restart"] = "write_restart"
        self._block_dict = block_dict

    def load_default(self, file_content=None):
        if file_content is None:
            file_content = (
                "units                 metal\n"
                + "dimension             3\n"
                + "boundary              p p p\n"
                + "atom_style            atomic\n"
                + "read_data             structure.inp\n"
                + "include               potential.inp\n"
                + "fix___ensemble        all nve\n"
                + "variable___dumptime   equal 100\n"
                + "variable___thermotime equal 100\n"
                + "dump___1              all custom ${dumptime} dump.out id type xsu ysu zsu fx fy fz vx vy vz\n"
                + "dump_modify___1       sort id format line \"%d %d %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g %20.15g\"\n"
                + "thermo_style          custom step temp pe etotal pxx pxy pxz pyy pyz pzz vol\n"
                + "thermo_modify         format float %20.15g\n"
                + "thermo                ${thermotime}\n"
                + "run                   0\n"
            )
        self.load_string(file_content)

    def pressure_to_lammps(self, pressure, rotation_matrix):
        """
        Convert a singular value, list of values, or combination of values to the appropriate six elements for Lammps
        pxx, pyy, pzz, pxy, pxz, and pyz pressure tensor representation.

        Lammps handles complex cells in a particular way, namely by using an upper triangular cell. This means we may
        need to convert our pressure tensor to a new coordinate frame. We also handle that transformation here.

        I case of a single pressure value, it is again returned as a single pressure value, to be used with the "iso"
        option (i.e., coupled deformation in x, y, and z).

        Finally, we also ensure that the units are converted from pyiron's GPa to whatever Lammps needs.

        Args:
            pressure (float/list/tuple/numpy.ndarray): The pressure(s) to convert.
            rotation_matrix (numpy.ndarray): The 3x3 matrix rotating from the pyiron to Lammps coordinate frame.

        Returns:
            (list): pxx, pyy, pzz, pxy, pxz, and pyz to be passed to Lammps.

            or

            (float): a single, isotropic pressure to be used with the "iso" option
        """

        # If pressure is a scalar, only unit conversion is needed.
        if np.isscalar(pressure):
            return float(pressure) * LAMMPS_UNIT_CONVERSIONS[self["units"]]["pressure"]

        # Normalize pressure to a list of 6 entries of either float or NoneType type.
        if len(pressure) > 6:
            raise ValueError("Pressure can have a maximum of 6 values, (x, y, z, xy, xz, and yz), but got " +
                             "{}".format(len(pressure)))
        pressure = [float(p) if p is not None else None
                    for p in pressure]
        pressure += (6 - len(pressure)) * [None]

        if all(p is None for p in pressure):
            raise ValueError("Pressure cannot have a length but all be None")

        # If necessary, rotate the pressure tensor to the Lammps coordinate frame.
        # Isotropic, hydrostatic pressures are rotation invariant.
        if not np.isclose(np.matrix.trace(rotation_matrix), 3) and not self._is_isotropic_hydrostatic(pressure):
            if any(p is None for p in pressure):
                raise ValueError("Cells which are not orthorhombic or an upper-triangular cell are incompatible with Lammps "
                                 "constant pressure calculations unless the entire pressure tensor is defined. "
                                 "The reason is that Lammps demands such cells be represented with an "
                                 "upper-triangular unit cell, thus a rotation between Lammps and pyiron coordinate "
                                 "frames is required; it is not possible to rotate the pressure tensor if any of "
                                 "its components is None.")
            pxx, pyy, pzz, pxy, pxz, pyz = pressure
            pressure_tensor = np.array(
                [[pxx, pxy, pxz],
                 [pxy, pyy, pyz],
                 [pxz, pyz, pzz]]
            )
            lammps_pressure_tensor = rotation_matrix.T @ pressure_tensor @ rotation_matrix
            pressure = list(lammps_pressure_tensor[[0, 1, 2, 0, 0, 1], [0, 1, 2, 1, 2, 2]])

        return [(p * LAMMPS_UNIT_CONVERSIONS[self["units"]]["pressure"]
                 if p is not None
                 else p)
                for p in pressure]

    @staticmethod
    def _is_isotropic_hydrostatic(pressure):
        axial_all_alike = None not in pressure[:3] and np.allclose(pressure[:3], pressure[0])
        shear_all_none = all(p is None for p in pressure[3:])
        shear_all_zero = None not in pressure[3:] and np.allclose(pressure[3:], 0)
        return axial_all_alike and (shear_all_none or shear_all_zero)

    def calc_minimize(
        self,
        ionic_energy_tolerance=0.0,
        ionic_force_tolerance=1e-4,
        max_iter=100000,
        pressure=None,
        n_print=100,
        style='cg',
        rotation_matrix=None
    ):
        """
        Sets parameters required for minimization.

        Args:
            ionic_energy_tolerance (float): If the magnitude of difference between energies of two consecutive steps is
                lower than or equal to `ionic_energy_tolerance`, the minimisation terminates. (Default is 0.0 eV.)
            ionic_force_tolerance (float): If the magnitude of the global force vector at a step is lower than or equal
                to `ionic_force_tolerance`, the minimisation terminates. (Default is 1e-4 eV/angstrom.)
            e_tol (float): Sam as ionic_energy_tolerance (deprecated)
            f_tol (float): Sam as ionic_force_tolerance (deprecated)
            max_iter (int): Maximum number of minimisation steps to carry out. If the minimisation converges before
                `max_iter` steps, terminate at the converged step. If the minimisation does not converge up to
                `max_iter` steps, terminate at the `max_iter` step. (Default is 100000.)
            pressure (None/float/numpy.ndarray/list): Target pressure. If set to None, an isochoric (constant V)
                calculation is performed. If set to a scalar, the shear of the cell and the ratio of the x, y, and
                z components is kept constant, while an isotropic, hydrostatic pressure is applied. A list of up to
                length 6 can be given to specify xx, yy, zz, xy, xz, and yz components of the pressure tensor,
                respectively. These values can mix floats and `None` to allow only certain degrees of cell freedom
                to change. (Default is None, run isochorically.)
            n_print (int): Write (dump or print) to the output file every n steps (Default: 100)
            style ('cg'/'sd'/other values from Lammps docs): The style of the numeric minimization, either conjugate
                gradient, steepest descent, or other keys permissible from the Lammps docs on 'min_style'. (Default
                is 'cg' -- conjugate gradient.)
            rotation_matrix (numpy.ndarray): The rotation matrix from the pyiron to Lammps coordinate frame.
        """
        # This docstring is a source for the calc_minimize method in pyiron.lammps.base.LammpsBase.calc_minimize and
        # pyiron.lammps.interactive.LammpsInteractive.calc_minimize -- Please ensure that changes to signature or
        # defaults stay consistent!

        max_evaluations = 100 * max_iter

        if self["units"] not in LAMMPS_UNIT_CONVERSIONS.keys():
            raise NotImplementedError
        energy_units = LAMMPS_UNIT_CONVERSIONS[self["units"]]["energy"]
        force_units = LAMMPS_UNIT_CONVERSIONS[self["units"]]["force"]

        ionic_energy_tolerance *= energy_units
        ionic_force_tolerance *= force_units

        if pressure is not None:
            pressure = self.pressure_to_lammps(pressure, rotation_matrix)
            if np.isscalar(pressure):
                str_press = " iso {}".format(pressure)
            else:
                str_press = ""
                for press, str_axis in zip(pressure, [" x ", " y ", " z ", " xy ", " xz ", " yz "]):
                    if press is not None:
                        str_press += str_axis + str(press)
                if len(str_press) > 1:
                    str_press += " couple none"
            self.set(fix___ensemble=r"all box/relax" + str_press)
        self.remove_keys(["fix___nve"])
        self.set(min_style=style)
        self.set(
            minimize=str(ionic_energy_tolerance)
            + " "
            + str(ionic_force_tolerance)
            + " "
            + str(int(max_iter))
            + " "
            + str(int(max_evaluations))
        )
        self.remove_keys(["run", "velocity"])
        self.modify(variable___dumptime="equal " + str(n_print), thermo=n_print)
        self.modify(variable___thermotime="equal " + str(n_print), thermo=n_print)

    def calc_static(self):
        self.set(run="0")
        self.remove_keys(["minimize", "velocity"])

    def set_initial_velocity(
        self,
        temperature,
        seed=None,
        gaussian=False,
        append_value=False,
        zero_lin_momentum=True,
        zero_rot_momentum=True,
        job_name="",
    ):
        """
        Create initial velocities via velocity all create. More information can be found on LAMMPS website:
        https://lammps.sandia.gov/doc/velocity.html

        Args:
            temperature: (int or float)
            seed: (int) Seed for the initial random number generator
            gaussian: (True/False) Create velocity according to the Gaussian distribution (otherwise uniform)
            append_value: (True/False) Add the velocity values to the current velocities (probably not functional now)
            zero_lin_momentum: (True/False) Cancel the total linear momentum
            zero_rot_momentum: (True/False) Cancel the total angular momentum
            job_name: (str) job name to generate seed
        """

        if seed is None:
            seed = self.generate_seed_from_job(job_name=job_name, seed=1)
        arg = ""
        if gaussian:
            arg = " dist gaussian"
        if append_value:
            arg += " sum yes"
        if not zero_lin_momentum:
            arg += " mom no"
        if not zero_rot_momentum:
            arg += " rot no"
        self.modify(
            velocity="all create " + str(temperature) + " " + str(seed) + arg,
            append_if_not_present=True,
        )

    @staticmethod
    def generate_seed_from_job(job_name="", seed=0):
        """
        Generate a unique seed from the job name.

        Args:
            job_name (str): job_name of the current job to generate the seed
            seed (int): integer to access part of the seed

        Returns:
            int: random seed generated based on the hash
        """
        return int(
            str(int(hashlib.sha256(job_name.encode()).hexdigest(), 16))[
                5 * seed : 5 * seed + 5
            ]
        )

    def calc_md(
        self,
        temperature=None,
        pressure=None,
        n_ionic_steps=1000,
        time_step=1.0,
        n_print=100,
        temperature_damping_timescale=100.0,
        pressure_damping_timescale=1000.0,
        seed=None,
        tloop=None,
        initial_temperature=None,
        langevin=False,
        delta_temp=None,
        delta_press=None,
        job_name="",
        rotation_matrix=None
    ):
        """
        Set an MD calculation within LAMMPS. Nosé Hoover is used by default.

        Args:
            temperature (None/float/list): Target temperature value(-s). If set to None, an NVE calculation is performed.
                                           It is required when the pressure is set or langevin is set
                                           It can be a list of temperature values, containing the initial target
                                           temperature and the final target temperature (in between the target value
                                           is varied linearly).
            pressure (None/float/numpy.ndarray/list): Target pressure. If set to None, an NVE or an NVT calculation is
                performed. If set to a scalar, the shear of the cell and the ratio of the x, y, and z components is kept
                constant, while an isotropic, hydrostatic pressure is applied. A list of up to length 6 can be given to
                specify xx, yy, zz, xy, xz, and yz components of the pressure tensor, respectively. These values can mix
                floats and `None` to allow only certain degrees of cell freedom to change. (Default is None, run
                isochorically.)
            n_ionic_steps (int): Number of ionic steps
            time_step (float): Step size in fs between two steps.
            n_print (int):  Print frequency
            temperature_damping_timescale (float): The time associated with the thermostat adjusting the temperature.
                                                   (In fs. After rescaling to appropriate time units, is equivalent to
                                                   Lammps' `Tdamp`.)
            pressure_damping_timescale (float): The time associated with the barostat adjusting the temperature.
                                                (In fs. After rescaling to appropriate time units, is equivalent to
                                                Lammps' `Pdamp`.)
            seed (int):  Seed for the random number generation (required for the velocity creation)
            tloop:
            initial_temperature (None/float):  Initial temperature according to which the initial velocity field
                                               is created. If None, the initial temperature will be twice the target
                                               temperature (which would go immediately down to the target temperature
                                               as described in equipartition theorem). If 0, the velocity field is not
                                               initialized (in which case  the initial velocity given in structure will
                                               be used). If any other number is given, this value is going to be used
                                               for the initial temperature.
            langevin (bool): (True or False) Activate Langevin dynamics
            delta_temp (float): Thermostat timescale, but in your Lammps time units, whatever those are. (DEPRECATED.)
            delta_press (float): Barostat timescale, but in your Lammps time units, whatever those are. (DEPRECATED.)
            job_name (str): Job name of the job to generate a unique random seed.
            rotation_matrix (numpy.ndarray): The rotation matrix from the pyiron to Lammps coordinate frame.
        """
        if self["units"] not in LAMMPS_UNIT_CONVERSIONS.keys():
            raise NotImplementedError
        time_units = LAMMPS_UNIT_CONVERSIONS[self["units"]]["time"]
        temperature_units = LAMMPS_UNIT_CONVERSIONS[self["units"]]["temperature"]

        # Transform time
        if time_step is not None:
            try:
                self["timestep"] = time_step * time_units
            except KeyError:
                raise NotImplementedError()

        # Transform thermostat strength (time)
        if delta_temp is not None:
            warnings.warn(
                "WARNING: `delta_temp` is deprecated, please use `temperature_damping_timescale`."
            )
            temperature_damping_timescale = delta_temp
        else:
            temperature_damping_timescale *= time_units

        # Transform barostat strength (time)
        if delta_press is not None:
            warnings.warn(
                "WARNING: `delta_press` is deprecated, please use `pressure_damping_timescale`."
            )
            pressure_damping_timescale = delta_press
        else:
            pressure_damping_timescale *= time_units

        # Transform temperature
        if temperature is not None:
            temperature = np.array([temperature], dtype=float).flatten()
            if len(temperature)==1:
                temperature = np.array(2*temperature.tolist())
            elif len(temperature) != 2:
                raise ValueError("At most two temperatures can be provided "
                                 "(for a linearly ramping target temperature), "
                                 "but got {}".format(len(temperature)))
            temperature *= temperature_units

        # Apply initial overheating (default uses the theorem of equipartition of energy between KE and PE)
        if initial_temperature is None and temperature is not None:
            initial_temperature = 2 * temperature[0]

        if seed is None:
            seed = self.generate_seed_from_job(job_name=job_name)
        if seed <= 0:
            raise ValueError("Seed must be a positive integer larger than 0")
            
        # Set thermodynamic ensemble
        if pressure is not None:  # NPT
            if temperature is None or temperature.min() <= 0:
                raise ValueError("Target temperature for fix nvt/npt/nph cannot be 0 or negative")

            pressure = self.pressure_to_lammps(pressure, rotation_matrix)

            if np.isscalar(pressure):
                pressure_string = " iso {0} {0} {1}".format(pressure, pressure_damping_timescale)
            else:
                pressure_string = ""
                for coord, value in zip(["x", "y", "z", "xy", "xz", "yz"], pressure):
                    if value is not None:
                        pressure_string += " {0} {1} {1} {2}".format(
                            coord, value, pressure_damping_timescale
                        )

            if langevin:  # NPT(Langevin)
                fix_ensemble_str = "all nph" + pressure_string
                self.modify(
                    fix___langevin="all langevin {0} {1} {2} {3} zero yes".format(
                        str(temperature[0]),
                        str(temperature[1]),
                        str(temperature_damping_timescale),
                        str(seed),
                    ),
                    append_if_not_present=True,
                )
            else:  # NPT(Nose-Hoover)
                fix_ensemble_str = "all npt temp {0} {1} {2}".format(
                    str(temperature[0]),
                    str(temperature[1]),
                    str(temperature_damping_timescale),
                )
                fix_ensemble_str += pressure_string
        elif temperature is not None:  # NVT
            if temperature.min() <= 0:
                raise ValueError("Target temperature for fix nvt/npt/nph cannot be 0 or negative")

            if langevin:  # NVT(Langevin)
                fix_ensemble_str = "all nve"
                self.modify(
                    fix___langevin="all langevin {0} {1} {2} {3} zero yes".format(
                        str(temperature[0]),
                        str(temperature[1]),
                        str(temperature_damping_timescale),
                        str(seed),
                    ),
                    append_if_not_present=True,
                )
            else:  # NVT(Nose-Hoover)
                fix_ensemble_str = "all nvt temp {0} {1} {2}".format(
                    str(temperature[0]),
                    str(temperature[1]),
                    str(temperature_damping_timescale),
                )
        else:  # NVE
            if langevin:
                warnings.warn("Temperature not set; Langevin ignored.")
            fix_ensemble_str = "all nve"
            initial_temperature = 0

        if tloop is not None:
            fix_ensemble_str += " tloop " + str(tloop)

        self.remove_keys(["minimize"])
        self.modify(
            fix___ensemble=fix_ensemble_str,
            variable___dumptime=" equal {} ".format(n_print),
            thermo=int(n_print),
            run=int(n_ionic_steps),
            append_if_not_present=True,
        )
        self.modify(
            fix___ensemble=fix_ensemble_str,
            variable___thermotime=" equal {} ".format(n_print),
            thermo=int(n_print),
            run=int(n_ionic_steps),
            append_if_not_present=True,
        )

        if initial_temperature > 0:
            self.set_initial_velocity(
                temperature=initial_temperature,
                seed=seed,
                gaussian=True,
                job_name=job_name
            )

    def calc_vcsgc(
        self,
        mu,
        ordered_element_list,
        target_concentration=None,
        kappa=1000.,
        mc_step_interval=100,
        swap_fraction=0.1,
        temperature_mc=None,
        window_size=None,
        window_moves=None,
        temperature=None,
        pressure=None,
        n_ionic_steps=1000,
        time_step=1.0,
        n_print=100,
        temperature_damping_timescale=100.0,
        pressure_damping_timescale=1000.0,
        seed=None,
        initial_temperature=None,
        langevin=False,
        job_name="",
        rotation_matrix=None
    ):
        """
        Run variance-constrained semi-grand-canonical MD/MC for a binary system. In addition to VC-SGC arguments, all
        arguments for a regular MD calculation are also accepted.

        https://vcsgc-lammps.materialsmodeling.org

        Note:
            For easy visualization later (with `get_structure`), it is highly recommended that the initial structure
            contain at least one atom of each species.

        Warning:
            Assumes the units are metal, otherwise units for the constraints may be off.

        Args:
            mu (dict): A dictionary of chemical potentials, one for each element the potential treats, where the
                dictionary keys are just the chemical symbol. Note that only the *relative* chemical potentials are used
                here, such that the swap acceptance probability is influenced by the chemical potential difference
                between the two species (a more negative value increases the odds of swapping *to* that element.)
            ordered_element_list (list): A list of the chemical species symbols in the order they appear in the
                definition of the potential in the Lammps' input file.
            target_concentration: A dictionary of target simulation domain concentrations for each species *in the
                potential*. Dictionary keys should be the chemical symbol of the corresponding species, and the sum of
                all concentrations must be 1. (Default is None, which runs regular semi-grand-canonical MD/MC without
                any variance constraint.)
            kappa: Variance constraint for the MC. Larger value means a tighter adherence to the target concentrations.
                (Default is 1000.)
            mc_step_interval (int): How many steps of MD between each set of MC moves. (Default is 100.) Must divide the
                number of ionic steps evenly.
            swap_fraction (float): The fraction of atoms whose species is swapped at each MC phase. (Default is 0.1.)
            temperature_mc (float): The temperature for accepting MC steps. (Default is None, which uses the MD
                temperature.)
            window_size (float): The size of the sampling window for parallel calculations as a fraction of something
                unspecified in the VC-SGC docs, but it must lie between 0.5 and 1. (Default is None, window is
                determined automatically.)
            window_moves (int): The number of times the sampling window is moved during one MC cycle. (Default is None,
                number of moves is determined automatically.)
            rotation_matrix (numpy.ndarray): The rotation matrix from the pyiron to Lammps coordinate frame.
        """
        self.calc_md(
            temperature=temperature,
            pressure=pressure,
            n_ionic_steps=n_ionic_steps,
            time_step=time_step,
            n_print=n_print,
            temperature_damping_timescale=temperature_damping_timescale,
            pressure_damping_timescale=pressure_damping_timescale,
            seed=seed,
            tloop=None,
            initial_temperature=initial_temperature,
            langevin=langevin,
            job_name=job_name,
            rotation_matrix=rotation_matrix
        )

        if self["units"] not in LAMMPS_UNIT_CONVERSIONS.keys():
            raise NotImplementedError
        temperature_units = LAMMPS_UNIT_CONVERSIONS[self["units"]]["temperature"]
        energy_units = LAMMPS_UNIT_CONVERSIONS[self["units"]]["energy"]

        if temperature_mc is None:
            if temperature is None:
                raise ValueError("If temperature is not given, temperature_mc must be.")
            temperature_mc = temperature * temperature_units

        if seed is None:
            seed = self.generate_seed_from_job(job_name=job_name)

        if set(mu.keys()) != set(ordered_element_list):
            raise ValueError("Exactly one chemical potential must be given for each element treated by the potential.")

        calibrating_el = ordered_element_list[0]
        # Apply the actual SGC string
        fix_vcsgc_str = "all sgcmc {0} {1} {2} {3} randseed {4}".format(
            str(mc_step_interval),
            str(swap_fraction),
            str(temperature_mc),
            str(" ".join(str((mu[el] - mu[calibrating_el]) * energy_units) for el in ordered_element_list[1:])),
            str(seed),
        )

        # Add VC to the SGC if target concentrations were provided
        if target_concentration is not None:
            if set(target_concentration.keys()) != set(ordered_element_list):
                raise ValueError("Exactly one target concentration must be given for each element treated by the "
                                 "potential.")

            if not np.isclose(np.sum([v for _, v in target_concentration.items()]), 1):
                raise ValueError("Target concentrations must sum to 1 but were {}.".format(
                    np.sum([v for _, v in target_concentration.items()])
                ))

            fix_vcsgc_str += " variance {0} {1}".format(
                str(kappa),
                str(" ".join([str(target_concentration[el]) for el in ordered_element_list[1:]]))
            )

        # Set optional windowing parameters
        if window_moves is not None:
            if int(window_moves) != window_moves or window_moves < 0:
                raise ValueError("Window moves must be a non-negative integer.")
            fix_vcsgc_str += " window_moves {0}".format(window_moves)
        if window_size is not None:
            if not 0.5 <= window_size <= 1.0:
                raise ValueError("Window size must be a fraction between 0.5 and 1")
            fix_vcsgc_str += " window_size {0}".format(window_size)

        self.modify(
            fix___vcsgc=fix_vcsgc_str,
            append_if_not_present=True
        )

    def measure_mean_value(self, key, every=1, repeat=None, name=None, atom=False):
        """
            Args:
                key (str): property to take an average value of (e.g. 'energy_pot' v.i.)
                every (int): number of steps there should be between two measurements
                repeat (int): number of measurements for each output (default: n_print/every)
                name (str): name to give in the output string (ignored if a pyiron predefined tag is used)

            Comments:
                Currently available keys: 'energy_pot', 'energy_tot', 'temperature', 'volume',
                                          'pressures', 'positions', 'forces, 'velocities'
                Future keys: 'cells'
        """

        if every<=0:
            raise AssertionError('every must be a positive integer')
        if repeat is None:
            self['variable___mean_repeat_times'] = 'equal round(${thermotime}/'+str(every)+')'
        else:
            self['variable___mean_repeat_times'] = 'equal {}'.format(repeat)
        if key=='energy_pot':
            self._measure_mean_value('energy_pot', 'pe', every)
        elif key=='energy_tot':
            self._measure_mean_value('energy_tot', 'etotal', every)
        elif key=='temperature':
            self._measure_mean_value('temperature', 'temp', every)
        elif key=='volume':
            self._measure_mean_value('volume', 'vol', every)
        elif key=='pressures':
            self._measure_mean_value('pressure', ['pxx', 'pyy', 'pzz', 'pxy', 'pxz', 'pyz'], every)
        elif key=='positions':
            self['compute___unwrap'] = 'all property/atom xu yu zu'
            self['fix___mean_positions'] = ('all ave/atom '
                                            +str(every)
                                            +' ${mean_repeat_times} ${thermotime} c_unwrap[*]'
                                           )
            self['dump___1'] = self['dump___1']+' '+' '.join(['f_mean_positions[{}]'.format(ii+1) for ii in range(3)])
            self['dump_modify___1'] = self['dump_modify___1'][:-1]+' '+' '.join(['%20.15g']*3)+'"'
        elif key=='forces':
            self._measure_mean_value('forces', ['fx', 'fy', 'fz'], every, atom=True)
        elif key=='velocities':
            self._measure_mean_value('velocities', ['vx', 'vy', 'vz'], every, atom=True)
        elif name is not None:
            if '**' in key:
                warnings.warn('** is replaced by ^ (as it is understood by LAMMPS)')
                key = key.replace('**', '^')
            self._measure_mean_value(name, key, every, atom)
        else:
            raise NotImplementedError(key+' is not implemented')

    def energy_pot_per_atom(self):
        if self['compute___energy_pot_per_atom'] is None:
            self['compute___energy_pot_per_atom'] = 'all pe/atom'
            self['dump___1'] += ' c_energy_pot_per_atom'
            self['dump_modify___1'] = self['dump_modify___1'][:-1] + ' %20.15g"'

    def _measure_mean_value(self, key_pyiron, key_lmp, every, atom=False):
        """
            Args:
                key (str): property to take an average value of (e.g. 'energy_pot' v.i.)
                every (int): number of steps there should be between two measurements
                repeat (int): number of measurements for each output (default: n_print/every)
                freq (int): output frequency (default: n_print)

            Comments:
                Currently available keys: 'energy_pot', 'energy_tot', 'temperature', 'volume',
                                          'pressures', 'positions', 'forces, 'velocities'
                Future keys: 'cells'
        """
        if isinstance(key_lmp, str):
            self['variable___{}'.format(key_pyiron)] = 'equal {}'.format(key_lmp)
            self['fix___mean_{}'.format(key_pyiron)] = ('all ave/time '
                                                        +str(every)
                                                        +' ${mean_repeat_times} ${thermotime} v_'
                                                        +str(key_pyiron))
            self['thermo_style'] = self['thermo_style']+' f_mean_{}'.format(key_pyiron)
        else:
            if atom is True:
                for ii, _ in enumerate(key_lmp):
                    self['variable___{}_{}'.format(key_pyiron, ii)] = 'atom {}'.format(key_lmp[ii])
                self['fix___mean_{}'.format(key_pyiron)] = ('all ave/atom '
                                                            +str(every)+
                                                            ' ${mean_repeat_times} ${thermotime} '
                                                            +str(' '.join(['v_{}_{}'.format(key_pyiron, ii) for ii in range(len(key_lmp))]))
                                                           )
                self['dump___1'] = self['dump___1']+' '+' '.join(['f_mean_{}[{}]'.format(key_pyiron, ii+1) for ii in range(len(key_lmp))])
                self['dump_modify___1'] = self['dump_modify___1'][:-1]+' '+' '.join(['%20.15g']*len(key_lmp))+'"'
            else:
                for ii, _ in enumerate(key_lmp):
                    self['variable___{}_{}'.format(key_pyiron, ii)] = 'equal {}'.format(key_lmp[ii])
                self['fix___mean_{}'.format(key_pyiron)] = ('all ave/time '
                                                            +str(every)
                                                            +' ${mean_repeat_times} ${thermotime} '
                                                            +str(' '.join(['v_{}_{}'.format(key_pyiron, ii) for ii in range(len(key_lmp))]))
                                                           )
                self['thermo_style'] = (self['thermo_style']
                                        +' '
                                        +' '.join(['f_mean_{}[{}]'.format(key_pyiron, ii+1) for ii in range(len(key_lmp))]))



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from sklearn import cluster
from pyiron.lammps.structure import UnfoldingPrism
from pyiron.atomistics.structure.atoms import pyiron_to_ase
import numpy as np
import pyscal.core as pc

__author__ = "Sarath Menon, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sarath Menon"
__email__ = "sarath.menon@rub.de"
__status__ = "development"
__date__ = "Nov 6, 2019"


def get_steinhardt_parameter_job(job, neighbor_method="cutoff", cutoff=0, n_clusters=2,
                                 q=(4, 6), averaged=False, clustering=True, iteration_step=-1):
    """
    Calculate Steinhardts parameters

    Args:
        job (job): pyiron job
        neighbor_method (str) : can be ['cutoff', 'voronoi']
        cutoff (float) : can be 0 for adaptive cutoff or any other value
        n_clusters (int) : number of clusters for K means clustering
        q (list) : can be from 2-12, the required q values to be calculated
        averaged (bool) : If True, calculates the averaged versions of the parameter
        clustering (bool) : If True, cluster based on the q values
        iteration_step (int) : The step at which the configuration is to be used

    Returns:
        q (list) : calculated q parameters
    """
    return get_steinhardt_parameter_structure(
        structure=job.get_structure(iteration_step=iteration_step),
        neighbor_method=neighbor_method,
        cutoff=cutoff,
        n_clusters=n_clusters,
        q=q,
        averaged=averaged,
        clustering=cluster
    )


def get_steinhardt_parameter_structure(structure, neighbor_method="cutoff", cutoff=0, n_clusters=2,
                                       q=(4, 6), averaged=False, clustering=True):
    """
    Calculate Steinhardts parameters

    Args:
        job (job): pyiron job
        neighbor_method (str) : can be ['cutoff', 'voronoi']
        cutoff (float) : can be 0 for adaptive cutoff or any other value
        n_clusters (int) : number of clusters for K means clustering
        q (list) : can be from 2-12, the required q values to be calculated
        averaged (bool) : If True, calculates the averaged versions of the parameter
        clustering (bool) : If True, cluster based on the q values

    Returns:
        q (list) : calculated q parameters

    """
    sys = pc.System()
    sys.read_inputfile(
        pyiron_to_ase(structure),
        format='ase'
    )

    sys.find_neighbors(
        method=neighbor_method,
        cutoff=cutoff
    )

    sys.calculate_q(
        q,
        averaged=averaged
    )

    sysq = sys.get_qvals(
        q,
        averaged=averaged
    )

    if clustering:
        cl = cluster.KMeans(
            n_clusters=n_clusters
        )

        ind = cl.fit(list(zip(*sysq))).labels_ == 0
        return sysq, ind
    else:
        return sysq

def analyse_centro_symmetry(atoms, num_neighbors=12):
    """
    Analyse centrosymmetry parameter

    Args:
        atoms: Atoms object
        num_neighbors (int) : number of neighbors

    Returns:
        csm (list) : list of centrosymmetry parameter
    """
    sys = pc.System()
    sys.read_inputfile(atoms, format="ase")
    sys.calculate_centrosymmetry(nmax=num_neighbors)
    atoms = sys.atoms
    return np.array([atom.centrosymmetry for atom in atoms])

def analyse_cna_adaptive(atoms, mode="total"):
    """
    Use common neighbor analysis

    Args:
        atoms (pyiron.structure.atoms.Atoms): The structure to analyze.
        mode ("total"/"numeric"/"str"): Controls the style and level
        of detail of the output.
        total : return number of atoms belonging to each structure
        numeric : return a per atom list of numbers- 0 for unknown,
            1 fcc, 2 hcp, 3 bcc and 4 icosa
        str : return a per atom string of sructures

    Returns:
        (depends on `mode`)
    """
    if not mode in ["total", "numeric", "str"]:
        raise ValueError("Unsupported mode")

    sys = pc.System()
    sys.read_inputfile(atoms, format="ase")
    cna = sys.calculate_cna(cutoff=None)

    if mode == "total":
        return cna
    else:
        atoms = sys.atoms
        cnalist = ([atom.structure for atom in atoms])
        if mode == "numeric":
            return cnalist
        else:
            dd = ["unknown", "fcc", "hcp", "bcc", "ico"]
            cnalist = [dd[int(x)] for x in cnalist]
            return cnalist

def analyse_voronoi_volume(atoms):
    """
    Calculate the Voronoi volume of atoms

    Args:
        atoms : (pyiron.structure.atoms.Atoms): The structure to analyze.
    """
    sys = pc.System()
    sys.read_inputfile(atoms, format="ase")
    sys.find_neighbors(method="voronoi")
    atoms = sys.atoms
    vols = np.array([atom.volume for atom in atoms])
    return vols



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from ase.build import bulk
from ase.build import (
        add_adsorbate,
        add_vacuum,
        bcc100,
        bcc110,
        bcc111,
        diamond100,
        diamond111,
        fcc100,
        fcc110,
        fcc111,
        fcc211,
        hcp0001,
        hcp10m10,
        mx2,
        hcp0001_root,
        fcc111_root,
        bcc111_root,
        root_surface,
        root_surface_analysis,
        surface as ase_surf,
    )
import numpy as np
from pyiron.atomistics.structure.pyironase import publication as publication_ase
from pyiron.atomistics.structure.atoms import CrystalStructure, ase_to_pyiron
from pyiron_base import Settings
import types

__author__ = "Sudarsan Surendralal"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "May 1, 2020"

s = Settings()


def create_surface(
    element, surface_type, size=(1, 1, 1), vacuum=1.0, center=False, pbc=True, **kwargs
):
    """
    Generate a surface based on the ase.build.surface module.

    Args:
        element (str): Element name
        surface_type (str): The string specifying the surface type generators available through ase (fcc111,
        hcp0001 etc.)
        size (tuple): Size of the surface
        vacuum (float): Length of vacuum layer added to the surface along the z direction
        center (bool): Tells if the surface layers have to be at the center or at one end along the z-direction
        pbc (list/numpy.ndarray): List of booleans specifying the periodic boundary conditions along all three
                                  directions.
        **kwargs: Additional, arguments you would normally pass to the structure generator like 'a', 'b',
        'orthogonal' etc.

    Returns:
        pyiron.atomistics.structure.atoms.Atoms instance: Required surface

    """
    # https://gitlab.com/ase/ase/blob/master/ase/lattice/surface.py
    if pbc is None:
        pbc = True
    s.publication_add(publication_ase())
    for surface_class in [
        add_adsorbate,
        add_vacuum,
        bcc100,
        bcc110,
        bcc111,
        diamond100,
        diamond111,
        fcc100,
        fcc110,
        fcc111,
        fcc211,
        hcp0001,
        hcp10m10,
        mx2,
        hcp0001_root,
        fcc111_root,
        bcc111_root,
        root_surface,
        root_surface_analysis,
        ase_surf,
    ]:
        if surface_type == surface_class.__name__:
            surface_type = surface_class
            break
    if isinstance(surface_type, types.FunctionType):
        if center:
            surface = surface_type(
                symbol=element, size=size, vacuum=vacuum, **kwargs
            )
        else:
            surface = surface_type(symbol=element, size=size, **kwargs)
            z_max = np.max(surface.positions[:, 2])
            surface.cell[2, 2] = z_max + vacuum
        surface.pbc = pbc
        return ase_to_pyiron(surface)
    else:
        return None


def create_hkl_surface(lattice, hkl, layers, vacuum=1.0, center=False, pbc=True):
    """
    Use ase.build.surface to build a surface with surface normal (hkl).

    Args:
        lattice (pyiron.atomistics.structure.atoms.Atoms/str): bulk Atoms
            instance or str, e.g. "Fe", from which to build the surface
        hkl (list): miller indices of surface to be created
        layers (int): # of atomic layers in the surface
        vacuum (float): vacuum spacing
        center (bool): shift all positions to center the surface
            in the cell

    Returns:
        pyiron.atomistics.structure.atoms.Atoms instance: Required surface
    """
    # https://gitlab.com/ase/ase/blob/master/ase/lattice/surface.py
    s.publication_add(publication_ase())

    surface = ase_surf(lattice, hkl, layers)
    z_max = np.max(surface.positions[:, 2])
    surface.cell[2, 2] = z_max + vacuum
    if center:
        surface.positions += 0.5 * surface.cell[2] - [0, 0, z_max/2]
    surface.pbc = pbc
    return ase_to_pyiron(surface)


def create_structure(element, bravais_basis, lattice_constant):
    """
    Create a crystal structure using pyiron's native crystal structure generator

    Args:
        element (str): Element name
        bravais_basis (str): Basis type
        lattice_constant (float/list): Lattice constants

    Returns:
        pyiron.atomistics.structure.atoms.Atoms: The required crystal structure

    """
    return CrystalStructure(
        element=element,
        bravais_basis=bravais_basis,
        lattice_constants=[lattice_constant],
    )


def create_ase_bulk(
    name,
    crystalstructure=None,
    a=None,
    c=None,
    covera=None,
    u=None,
    orthorhombic=False,
    cubic=False,
):
    """
    Creating bulk systems using ASE bulk module. Crystal structure and lattice constant(s) will be guessed if not
    provided.

    name (str): Chemical symbol or symbols as in 'MgO' or 'NaCl'.
    crystalstructure (str): Must be one of sc, fcc, bcc, hcp, diamond, zincblende,
                            rocksalt, cesiumchloride, fluorite or wurtzite.
    a (float): Lattice constant.
    c (float): Lattice constant.
    c_over_a (float): c/a ratio used for hcp.  Default is ideal ratio: sqrt(8/3).
    u (float): Internal coordinate for Wurtzite structure.
    orthorhombic (bool): Construct orthorhombic unit cell instead of primitive cell which is the default.
    cubic (bool): Construct cubic unit cell if possible.

    Returns:

        pyiron.atomistics.structure.atoms.Atoms: Required bulk structure
    """
    s.publication_add(publication_ase())
    return ase_to_pyiron(bulk(
        name=name,
        crystalstructure=crystalstructure,
        a=a,
        c=c,
        covera=covera,
        u=u,
        orthorhombic=orthorhombic,
        cubic=cubic,
    ))



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from pyiron.vasp.vasp import Vasp

__author__ = "Jan Janssen"
__copyright__ = "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - " \
                "Computational Materials Design (CM) Department"
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2018"


class VaspSol(Vasp):
    def __init__(self, project, job_name):
        super(VaspSol, self).__init__(project, job_name)
        self.__name__ = "VaspSol"
        self.__version__ = None  # Reset the version number to the executable is set automatically
        self._executable = None
        self._executable_activate()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from collections import OrderedDict
import numpy as np
import scipy.constants
from pyiron.atomistics.structure.atoms import Atoms
from pyiron.atomistics.structure.periodic_table import PeriodicTable

__author__ = "Sudarsan Surendralal, Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Sudarsan Surendralal"
__email__ = "surendralal@mpie.de"
__status__ = "production"
__date__ = "Feb 4, 2018"

BOHR_TO_ANGSTROM = (
    scipy.constants.physical_constants["Bohr radius"][0] / scipy.constants.angstrom
)


def read_atoms(filename="structure.sx"):
    """
    Args:
        filename (str): Filename of the sphinx structure file

    Returns:
        pyiron.objects.structure.atoms.Atoms instance

    """
    file_string = []
    with open(filename) as f:
        for line in f:
            line = line.strip()
            file_string.append(line)
    cell_trigger = "cell"
    cell_string = list()
    species_list = list()
    species_trigger = "element"
    positions_dict = OrderedDict()
    positions = list()
    pse = PeriodicTable()
    for i, line in enumerate(file_string):
        if cell_trigger in line:
            for j in range(len(file_string)):
                line_str = file_string[i + j]
                cell_string.append(line_str)
                if ";" in line_str:
                    break
        if species_trigger in line:
            species = (
                line.strip().split("=")[-1].replace(";", "").replace('"', "").strip()
            )
            species_list.append(pse.element(species))
            positions_dict[species] = 0
            for j in range(len(file_string) - i):
                line_str = file_string[i + j]
                k = 0
                if "atom" in line_str:
                    break_loop = False
                    while not break_loop:
                        position_string = " ".join(
                            file_string[i + j + k].split("=")[-1]
                        )
                        replace_list = ["[", "]", ";", "}",
                            "movable", "X", "Y", "Z"]
                        for rep in replace_list:
                            position_string = (
                                "".join(position_string).replace(rep, " ").split()
                            )
                        positions.append(
                            np.array(position_string[0].split(","), dtype=float)
                        )
                        positions_dict[species] += 1
                        k += 1
                        if (i + j + k) <= len(file_string) - 1:
                            if (
                                "element" in file_string[i + j + k]
                                or "atom" not in file_string[i + j + k]
                            ):
                                break_loop = True
                    break
    indices = list()
    for i, val in enumerate(positions_dict.values()):
        indices.append(np.ones(val, dtype=int) * i)
    indices = np.hstack(indices)
    replace_list = ["cell", "=", "[", "]", ",", ";"]
    for rep in replace_list:
        cell_string = " ".join(cell_string).replace(rep, " ").split()
    cell = np.array(cell_string, dtype=float).reshape((3, 3)) * BOHR_TO_ANGSTROM
    atoms = Atoms(
        species=species_list,
        indices=indices,
        cell=cell,
        positions=np.array(positions) * BOHR_TO_ANGSTROM,
    )
    return atoms



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

from __future__ import print_function
from pyiron.atomistics.master.parallel import AtomisticParallelMaster
from pyiron_base import JobGenerator
import numpy as np

try:
    import pylab as plt
except (ImportError, RuntimeError):
    try:
        import matplotlib.pyplot as plt
    except (ImportError, RuntimeError):
        pass

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"


class KpointConvergenceJobGenerator(JobGenerator):
    @property
    def parameter_list(self):
        """

        Returns:
            (list)
        """
        return [
            kpoint
            for kpoint in range(
                self.input["min"],
                self.input["max"] + self.input["steps"],
                self.input["steps"],
            )
        ]

    @staticmethod
    def job_name(parameter):
        return "kpoint_mesh_" + str(parameter)

    @staticmethod
    def modify_job(job, parameter):
        job.set_kpoints(mesh=[parameter, parameter, parameter])
        return job


# ToDo: not all abstract methods implemented
class ConvKpointParallel(AtomisticParallelMaster):
    def __init__(self, project, job_name="encut_conv"):
        """

        Args:
            project:
            job_name:
        """
        super(ConvKpointParallel, self).__init__(project, job_name)
        self.__name__ = "ConvKpointParallel"
        self.__version__ = "0.0.1"

        # define default input
        self.input["steps"] = (2, "increase of kpoints")
        self.input["min"] = (4, "Kpoint Minimum")
        self.input["max"] = (8, "Kpoint Maximum")
        self._job_generator = KpointConvergenceJobGenerator(self)

    def write_input(self):
        self.input["num_points"] = len(
            range(
                self.input["min"],
                self.input["max"] + self.input["steps"],
                self.input["steps"],
            )
        )
        super(ConvKpointParallel, self).write_input()

    def collect_output(self):
        eng_lst, kpoint_lst = [], []
        for job_id in self.child_ids:  # add iter_jobs (should behave like a project)
            ham = self.project_hdf5.inspect(job_id)
            print("job_id: ", job_id, ham.status)
            eng_lst.append(ham["output/generic/energy_tot"][-1])
            kpoint_lst.append(int(ham.job_name.split("_")[-1]))
        arg_lst = np.argsort(kpoint_lst)
        self._output["energy"] = eng_lst[arg_lst]
        self._output["kpoints"] = kpoint_lst[arg_lst]

        with self.project_hdf5.open("output") as hdf5_out:
            for key, val in self._output.items():
                hdf5_out[key] = val

    def plot(self, plt_show=True):
        df = self.output_to_pandas()
        kpoint_lst, eng_lst = df["kpoints"], df["energy"]
        plt.plot(kpoint_lst, eng_lst, "x-", markersize=20)
        plt.xlabel("Kpoint Mesh")
        plt.ylabel("energy (eV)")
        if plt_show:
            plt.show()



# coding: utf-8
# Copyright (c) Max-Planck-Institut für Eisenforschung GmbH - Computational Materials Design (CM) Department
# Distributed under the terms of "New BSD License", see the LICENSE file.

import numpy as np
from pyiron_base import Settings
from ovito import ObjectNode
from ovito.data import Bonds, DataCollection
from ovito.modifiers import (
    CreateBondsModifier,
    CommonNeighborAnalysisModifier,
    CentroSymmetryModifier,
    VoronoiAnalysisModifier,
)

__author__ = "Jan Janssen"
__copyright__ = (
    "Copyright 2020, Max-Planck-Institut für Eisenforschung GmbH - "
    "Computational Materials Design (CM) Department"
)
__version__ = "1.0"
__maintainer__ = "Jan Janssen"
__email__ = "janssen@mpie.de"
__status__ = "development"
__date__ = "Sep 1, 2017"

s = Settings()


def analyse_ovito_cna_adaptive(atoms, mode="total"):
    """
    Use Ovito's common neighbor analysis binding.

    Args:
        atoms (pyrion.structure.atoms.Atoms): The structure to analyze.
        mode ("total"/"numeric"/"str"): Controls the style and level of detail of the output. (Default is "total", only
            return a summary of the values in the structure.)

    Returns:
        (depends on `mode`)
    """
    s.publication_add(publication())
    if not mode in ["total", "numeric", "str"]:
        raise ValueError("Unsupported mode")
    data = DataCollection.create_from_ase_atoms(atoms.copy())
    node = ObjectNode()
    node.source = data
    node.modifiers.append(
        CommonNeighborAnalysisModifier(
            mode=CommonNeighborAnalysisModifier.Mode.AdaptiveCutoff
        )
    )
    output = node.compute()
    if mode == "total":
        return output.attributes
    else:
        atoms_output = output.to_ase_atoms()
        if mode == "numeric":
            return atoms_output.get_array("Structure Type")
        elif mode == "str":
            cna_property = output.particle_properties.structure_type
            return np.array(
                [
                    cna_property.get_type_by_id(cnatype).name
                    for cnatype in atoms_output.get_array("Structure Type")
                ]
            )


def analyse_ovito_centro_symmetry(atoms, num_neighbors=12):
    """

    Args:
        atoms:
        num_neighbors:

    Returns:

    """
    s.publication_add(publication())
    data = DataCollection.create_from_ase_atoms(atoms.copy())
    node = ObjectNode()
    node.source = data
    node.modifiers.append(CentroSymmetryModifier(num_neighbors=num_neighbors))
    output = node.compute()
    return output.particle_properties["Centrosymmetry"].array


def analyse_ovito_diamond(atoms, mode="total"):
    """
    Use Ovito's Diamond structure analysis bindings.

    Args:
        atoms (pyrion.structure.atoms.Atoms): The structure to analyze.
        mode ("total"/"numeric"/"str"): Controls the style and level of detail of the output. (Default is "total", only
            return a summary of the values in the structure.)
    Returns:
        (depends on `mode`)
    """
    s.publication_add(publication())
    if not mode in ["total", "numeric", "str"]:
        raise ValueError("Unsupported mode")
    data = DataCollection.create_from_ase_atoms(atoms.copy())
    node = ObjectNode()
    node.source = data
    node.modifiers.append(
        IdentifyDiamondModifier()
    )
    output = node.compute()
    if mode == "total":
        return output.attributes
    else:
        atoms_output = output.to_ase_atoms()
        if mode == "numeric":
            return atoms_output.get_array("Structure Type")
        elif mode == "str":
            diamond_property = output.particle_properties.structure_type
            return np.array(
                [
                    diamond_property.get_type_by_id(cnatype).name
                    for cnatype in atoms_output.get_array("Structure Type")
                ]
            )


def analyse_ovito_voronoi_volume(atoms):
    """

    Args:
        atoms:

    Returns:

    """
    s.publication_add(publication())
    data = DataCollection.create_from_ase_atoms(atoms.copy())
    node = ObjectNode()
    node.source = data
    node.modifiers.append(VoronoiAnalysisModifier())
    output = node.compute()
    return output.particle_properties["Atomic Volume"].array


def publication():
    return {
        "ovito": {
            "Stukowski_2009": {
                "doi": "10.1088/0965-0393/18/1/015012",
                "url": "https://doi.org/10.1088%2F0965-0393%2F18%2F1%2F015012",
                "year": "2009",
                "month": "dec",
                "publisher": "{IOP} Publishing",
                "volume": "18",
                "number": "1",
                "pages": "015012",
                "author": ["Alexander Stukowski"],
                "title": "Visualization and analysis of atomistic simulation data with "
                "{OVITO}{\textendash}the Open Visualization Tool",
                "journal": "Modelling and Simulation in Materials Science and Engineering",
                "abstract": "The Open Visualization Tool (OVITO) is a new 3D visualization "
                "software designed for post-processing atomistic data obtained "
                "from molecular dynamics or Monte Carlo simulations. Unique "
                "analysis, editing and animations functions are integrated into "
                "its easy-to-use graphical user interface. The software is "
                "written in object-oriented C++, controllable via Python scripts "
                "and easily extendable through a plug-in interface. It is "
                "distributed as open-source software and can be downloaded "
                "from the website http://ovito.sourceforge.net/.",
            }
        }
    }
